#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# generated by datamodel-codegen:
#   filename:  colormap.yaml

from __future__ import annotations

from datetime import datetime
from enum import Enum
from uuid import UUID

from pydantic import (
    AnyUrl,
    Field,
    RootModel,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)

from .._model_config import CustomBaseModel


class AttributeAssociationData(CustomBaseModel):
    attribute_id: StrictStr = Field(..., title="Attribute Id")
    colormap_id: UUID = Field(..., title="Colormap Id")


class AttributeAssociationsData(CustomBaseModel):
    associations: list[AttributeAssociationData] = Field(..., max_length=128, min_length=1, title="Associations")


class AttributeColormapAssociation(CustomBaseModel):
    attribute_id: StrictStr = Field(..., title="Attribute ID")
    colormap_id: UUID = Field(..., title="Colormap ID")
    colormap_uri: AnyUrl = Field(..., title="Colormap URI")
    created_at: datetime = Field(..., title="Created at")
    created_by: UUID = Field(..., title="Created by")
    id: UUID = Field(..., title="ID")
    modified_at: datetime = Field(..., title="Modified at")
    modified_by: UUID = Field(..., title="Modified by")
    self_link: AnyUrl = Field(..., title="Self link")
    workspace_id: UUID = Field(..., title="Workspace ID")


class AttributeColormapAssociationList(CustomBaseModel):
    associations: list[AttributeColormapAssociation] = Field(..., title="Associations")


class ColorItem(RootModel[StrictInt]):
    root: StrictInt


class Color(RootModel[list[ColorItem]]):
    root: list[ColorItem] = Field(..., max_length=3, min_length=3)


class ColormapTypeEnum(Enum):
    continuous = "continuous"
    category = "category"
    discrete = "discrete"


class ContinuousColormapData(CustomBaseModel):
    attribute_controls: list[StrictFloat] = Field(..., min_length=2, title="Attribute Controls")
    colors: list[Color] = Field(..., max_length=1024, min_length=2, title="Colors")
    """
    An ordered list of RGB colors, containing between 2 and 1024 colors. Each color is represented by three integer values (for Red, Green, and Blue) ranging from 0 to 255.
    """
    dtype: ColormapTypeEnum = ColormapTypeEnum.continuous
    gradient_controls: list[StrictFloat] = Field(..., min_length=2, title="Gradient Controls")
    name: StrictStr = Field(..., title="Name")


class ContinuousColormapResponse(CustomBaseModel):
    attribute_controls: list[StrictFloat] = Field(..., min_length=1, title="Attribute controls")
    """
    List of ordered attribute values for gradient curve
    """
    colors: list[Color] = Field(..., max_length=1024, min_length=2, title="Colors")
    """
    An ordered list of RGB colors, containing between 2 and 1024 colors. Each color is represented by three integer values (for Red, Green, and Blue) ranging from 0 to 255.
    """
    created_at: datetime = Field(..., title="Created at")
    created_by: UUID = Field(..., title="Created by")
    gradient_controls: list[StrictFloat] = Field(..., min_length=1, title="Gradient controls")
    """
    List of gradient positions for gradient curve, floats between 0 and 1
    """
    id: UUID = Field(..., title="ID")
    modified_at: datetime = Field(..., title="Modified at")
    modified_by: UUID = Field(..., title="Modified by")
    name: StrictStr = Field(..., title="Name")
    schema_: ColormapTypeEnum = Field(..., alias="schema")
    self_link: AnyUrl = Field(..., title="Self link")


class DiscreteColormapData(CustomBaseModel):
    colors: list[Color] = Field(..., max_length=1024, min_length=1, title="Colors")
    """
    An ordered list of RGB colors, containing between 1 and 1024 colors. Each color is represented by three integer values (for Red, Green, and Blue) ranging from 0 to 255.
    """
    dtype: ColormapTypeEnum = ColormapTypeEnum.discrete
    end_inclusive: list[StrictBool] = Field(..., title="End Inclusive")
    end_points: list[StrictFloat] = Field(..., title="End Points")
    name: StrictStr = Field(..., title="Name")


class DiscreteColormapResponse(CustomBaseModel):
    colors: list[Color] = Field(..., max_length=1024, min_length=1, title="Colors")
    """
    An ordered list of RGB colors, containing between 1 and 1024 colors. Each color is represented by three integer values (for Red, Green, and Blue) ranging from 0 to 255.
    """
    created_at: datetime = Field(..., title="Created at")
    created_by: UUID = Field(..., title="Created by")
    end_inclusive: list[StrictBool] = Field(..., title="End inclusive")
    """
    Boolean array indicating whether to include the endpoint in the bucket or not
    """
    end_points: list[StrictFloat] = Field(..., title="End points")
    """
    List of colormap bucket endpoints, must be one fewer than the number of colors
    """
    id: UUID = Field(..., title="ID")
    modified_at: datetime = Field(..., title="Modified at")
    modified_by: UUID = Field(..., title="Modified by")
    name: StrictStr = Field(..., title="Name")
    schema_: ColormapTypeEnum = Field(..., alias="schema")
    self_link: AnyUrl = Field(..., title="Self link")


class CategoryColormapData(CustomBaseModel):
    colors: list[Color] = Field(..., max_length=10000, min_length=1, title="Colors")
    """
    An ordered list of RGB colors, containing between 1 and 1024 colors. Each color is represented by three integer values (for Red, Green, and Blue) ranging from 0 to 255.
    """
    dtype: ColormapTypeEnum = ColormapTypeEnum.category
    map: list[StrictStr] = Field(..., title="Map")
    name: StrictStr = Field(..., title="Name")


class CategoryColormapResponse(CustomBaseModel):
    colors: list[Color] = Field(..., max_length=10000, min_length=1, title="Colors")
    """
    An ordered list of RGB colors, containing between 1 and 10000 colors. Each color is represented by three integer values (for Red, Green, and Blue) ranging from 0 to 255.
    """
    created_at: datetime = Field(..., title="Created at")
    created_by: UUID = Field(..., title="Created by")
    id: UUID = Field(..., title="ID")
    map: list[StrictStr] = Field(..., title="Map")
    """
    String array mapping arbitrary strings to colors
    """
    modified_at: datetime = Field(..., title="Modified at")
    modified_by: UUID = Field(..., title="Modified by")
    name: StrictStr = Field(..., title="Name")
    schema_: ColormapTypeEnum = Field(..., alias="schema")
    self_link: AnyUrl = Field(..., title="Self link")


class ColormapData(RootModel[ContinuousColormapData | DiscreteColormapData | CategoryColormapData]):
    root: ContinuousColormapData | DiscreteColormapData | CategoryColormapData = Field(..., title="Colormap Data")


class ColormapResponseList(CustomBaseModel):
    colormaps: list[CategoryColormapResponse | DiscreteColormapResponse | ContinuousColormapResponse] = Field(
        ..., title="Colormaps"
    )


class ResponseGetColormapByID(
    RootModel[ContinuousColormapResponse | CategoryColormapResponse | DiscreteColormapResponse]
):
    root: ContinuousColormapResponse | CategoryColormapResponse | DiscreteColormapResponse = Field(  # noqa: E501
        ..., title="ResponseGetColormapByID"
    )


class ResponsePostColormap(RootModel[ContinuousColormapResponse | CategoryColormapResponse | DiscreteColormapResponse]):
    root: ContinuousColormapResponse | CategoryColormapResponse | DiscreteColormapResponse = Field(  # noqa: E501
        ..., title="ResponsePostColormap"
    )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

__all__ = [
    "Association",
    "AssociationMetadata",
    "CategoryColormap",
    "ColormapMetadata",
    "ContinuousColormap",
    "DiscreteColormap",
]

from dataclasses import dataclass
from datetime import datetime
from uuid import UUID

from evo.common import ResourceMetadata, ServiceUser


def _validate_colors(colors: list[list[int]], minimum: int = 2, maximum: int = 1024):
    if not minimum <= len(colors) <= maximum:
        raise ValueError("Colors must have between 2 and 1024 entries")

    for color in colors:
        if len(color) != 3 or any(not isinstance(value, int) or value < 0 or value > 255 for value in color):
            raise ValueError("Each color must have exactly 3 values, each in the range 0-255")


@dataclass(frozen=True, kw_only=True)
class ContinuousColormap:
    """Continuous colormap data."""

    colors: list[list[int]]
    """A list of colors. Each color is an RGB representation of the color as list of 3 integers in the range 0-255.
    The list must have between 2 and 1024 colors."""

    attribute_controls: list[float]

    gradient_controls: list[float]

    def __post_init__(self):
        _validate_colors(self.colors, minimum=2, maximum=1024)


@dataclass(frozen=True, kw_only=True)
class DiscreteColormap:
    """Discrete colormap data."""

    colors: list[list[int]]
    """A list of colors. Each color is an RGB representation of the color as list of 3 integers in the range 0-255.
    The list must have between 1 and 1024 colors."""

    end_inclusive: list[bool]

    end_points: list[float]

    def __post_init__(self):
        _validate_colors(self.colors, minimum=1, maximum=1024)


@dataclass(frozen=True, kw_only=True)
class CategoryColormap:
    """Category colormap data."""

    colors: list[list[int]]
    """A list of colors. Each color is an RGB representation of the color as list of 3 integers in the range 0-255.
    The list must have between 1 and 10000 colors."""

    map: list[str]

    def __post_init__(self):
        _validate_colors(self.colors, minimum=1, maximum=10000)


@dataclass(frozen=True, kw_only=True)
class ColormapMetadata(ResourceMetadata):
    """Colormap metadata."""

    modified_at: datetime
    """The last modified timestamp."""

    modified_by: ServiceUser
    """The user that last modified the colormap."""

    colormap: ContinuousColormap | DiscreteColormap | CategoryColormap

    self_link: str
    """The URL of the colormap."""

    def url(self) -> str:
        return self.self_link


@dataclass(frozen=True, kw_only=True)
class Association:
    """An association between an attribute and a colormap."""

    attribute_id: str
    """The ID of the attribute."""

    colormap_id: UUID
    """The ID of the colormap."""


@dataclass(frozen=True, kw_only=True)
class AssociationMetadata(ResourceMetadata):
    """Association metadata."""

    modified_at: datetime
    """The last modified timestamp."""

    modified_by: ServiceUser
    """The user that last modified the association."""

    self_link: str
    """The URL of the association."""

    colormap_id: UUID
    """The ID of the colormap."""

    attribute_id: str
    """The ID of the attribute."""

    object_id: UUID
    """The ID of the object."""

    colormap_uri: str
    """The URI of the colormap."""

    def url(self) -> str:
        return self.self_link



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Block Model API
=============


    The Block Model API provides the ability to manage and report on block models in your Evo workspaces. Enable your
    product with Evo connected workflows by integrating with the Seequent Block Model API.

    The Block Model API supports a range of sub-blocking options, and both full and partial updates to block models,
    including updates of specified columns and/or sub-volumes. Block models are versioned, and the service supports
    reporting on the material content of current or previous versions, and comparing the content between versions of
    a block model.

    For more information on the Block Model API, see [Overview](/docs/blockmodel/overview), or the API references here.


This code is generated from the OpenAPI specification for Block Model API.
API version: 1.16.1
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod  # noqa: F401

from ..models import *  # noqa: F403

__all__ = ["ColumnOperationsApi"]


class ColumnOperationsApi:
    """API client for the Column_operations endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def notify_upload_complete(
        self,
        job_id: str,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> UpdatedResponse:  # noqa: F405
        """Advise that the update file has finished uploading

        Signals that you have finished uploading the update file to the server blob storage indicated by the `upload_url`, and that the update can continue. This moves the `job_status` from `PENDING_UPLOAD` to `QUEUED`.  After this endpoint is called, the `job_url` should then be polled with a GET request to wait for the `job_status` field of the job to become either `FAILED` or `COMPLETE`. On workflow success, a new version of the block model will be created that will contain the changes specified in the update request.  Note that this endpoint still needs to be hit even when an update does not require a file upload.  A minimum workspace role of Editor is required to use this endpoint.

        :param job_id: ID of the job this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'ff0b8078-c7ec-41ab-87fb-f3eb32812b32'`
        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "job_id": job_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": UpdatedResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/jobs/{job_id}/uploaded",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def update_block_model_from_latest_version(
        self,
        workspace_id: str,
        org_id: str,
        bm_id: str,
        update_data_lite_input: UpdateDataLiteInput,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> UpdateWithUrl:  # noqa: F405
        """Start a block model data update

        Starts the workflow to update the latest version of the block model. On workflow success, a new version of the block model will be created that will contain the changes specified in the request.  When a user wishes to update the blocks of a block model, they need to hit multiple endpoints to complete the workflow. This includes:  1. An initial request; 2. An upload of data (for `new` and `update` operations); 3. And finally, a confirmation that the update job can proceed from the state `PENDING_UPLOAD` to `QUEUED`.  This endpoint is the initial request, where the server does a sanity check/validation on what the client is attempting to do.  With regards to the request body, - We require the `columns` field to be supplied, which is used to convey which operations the client wishes the service to perform on the block model during the update. The sub-fields of columns (or operations) `new`, `update`, `rename`, `update_metadata`, and `delete` specify the downstream behaviour on the specified columns during the update workflow. Note that `new`, `update`, and `delete` can be provided in concert, whilst `rename` can only be provided by itself. `update_metadata` can be provided along with `new`, `update`, and `delete`, but only when there are no columns being renamed (i.e. when only the unit ID is being updated). Some general rules/mechanics for all of these fields:     - The reserved system columns [`i`, `j`, `k`, `sidx`, `x`, `y`, `z`, `dx`, `dy`, `dz`, `version_id`] and any other model type specific sub-block index columns, cannot be targeted using any of the update operations.     - Any list of titles provided cannot contain duplicates.     - In all cases, unless specifically stated otherwise, columns that are targeted in one operation cannot be targeted in another. - There is an optional field `update_type`, which is used to determine which blocks are updated in the block model. The supported update types are `replace` and `merge`.     - If `update_type` is set to `replace`, the values of the columns provided in the update file are updated for all blocks within the block model. If a block is not provided in the update file, the column values for that block are set to `null`. The columns not provided in the update file are not affected.     - If `update_type` is set to `merge` (the default), only the blocks and columns that are provided in the update file are updated. The blocks and columns not provided in the update file are not affected. - Note that the optional field `geometry_change` can only be provided for sub-blocked models.     - If `geometry_change` is set to `false`, no new sub-blocks can be created nor destroyed, only sub-blocks that exist in the latest version can be updated.     - If `geometry_change` is set to `true`, new sub-blocks can be created and/or existing sub-blocks can be destroyed, but all columns must be provided. If `update_type` is set to `merge`, then all sub-blocks within a parent block must be provided. - For `new` and `update` operations, the details of the data file for upload can be specified using the optional `input_options` field.  If the request is successful, the server then sends back a pollable `job_url` and an `upload_url` (if required). The `upload_url` is where the data associated with a `new` or `update` operation should be uploaded to, which can be done with a PUT request. Note that the generated URL contains a SAS token, which is used to authenticate the upload. An upload can be started no later than 30 minutes after the generation of the SAS token.  With regards to the data file to be uploaded, - The file can either be in Parquet or CSV format. - The file needs to contain either IJK (block index) or XYZ (centroid coordinate) information for each block/row, along with the data for the columns specified in the request body, using the correct data type. - If both IJK and XYZ columns are supplied in the data file, the service will only make use of the IJK information.  A minimum workspace role of Editor is required to use this endpoint.

        :param workspace_id: ID of the workspace this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef'`
        :param org_id: ID of the organization this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'bf1a040c-8c58-4bc2-bec2-c5ae7de8bd84'`
        :param bm_id: ID of the block model this call is scoped to. Represented as a v4 UUID.
            Format: `uuid`
            Example: `'e3c277c2-edc6-4a7a-8380-251dd19231f2'`
        :param update_data_lite_input:
            Example: `endpoints.UpdateDataLiteInput()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
            "bm_id": bm_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "202": UpdateWithUrl,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.PATCH,
            resource_path="/blockmodel/orgs/{org_id}/workspaces/{workspace_id}/block-models/{bm_id}/blocks",
            path_params=_path_params,
            header_params=_header_params,
            body=update_data_lite_input,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import logging
import socket
import threading
import unittest
from collections.abc import Callable, Mapping
from functools import partial
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from typing import Any
from unittest import mock
from unittest.mock import patch
from urllib.parse import ParseResult, urlparse

from aiohttp import ClientResponse, ClientSession, ClientTimeout, TCPConnector, multipart
from parameterized.parameterized import parameterized

from evo.aio.transport import AioTransport
from evo.common import HTTPHeaderDict, HTTPResponse, RequestMethod
from evo.common.exceptions import ClientValueError, RetryError, TransportError
from evo.common.test_tools import long_test
from evo.common.utils.retry import BackoffLinear
from evo.logging import getLogger

getLogger("aio.transport").setLevel(logging.DEBUG)


class EchoHandler(BaseHTTPRequestHandler):
    """Simple HTTP server request handler for testing."""

    def __init__(self, mock_object: mock.Mock, *args: Any, **kwargs: Any) -> None:
        self.__mock_object = mock_object
        super().__init__(*args, **kwargs)

    def __do_status(self) -> bool:
        self.__mock_object(RequestMethod(self.command), self.path)
        match self.path:
            case "/test/retry":  # Trigger a retry.
                return False
            case _:  # Normal response.
                self.send_response(200)

        return True

    def __do_headers(self) -> None:
        for field, value in self.headers.items():
            self.send_header(field, value)
        self.end_headers()

    def __do_body(self) -> None:
        data = self.rfile.read(int(self.headers.get("Content-Length", 0)))
        self.wfile.write(data)

    def handle_one_request(self) -> None:
        """Handle a single HTTP request."""
        # This method is copied from the BaseHTTPRequestHandler class and modified for testing purposes.
        try:
            self.raw_requestline = self.rfile.readline(65537)
            if len(self.raw_requestline) > 65536:
                self.requestline = ""
                self.request_version = ""
                self.command = ""
                self.send_error(HTTPStatus.REQUEST_URI_TOO_LONG)
                return
            if not self.raw_requestline:
                self.close_connection = True
                return
            if not self.parse_request():
                # An error code has been sent, just exit
                return

            # Begin: Custom code.
            try:
                if self.__do_status():
                    self.__do_headers()
                    if self.command != "HEAD":
                        self.__do_body()
            except Exception as e:
                self.send_error(500, str(e))
            # End: Custom code.

            self.wfile.flush()  # actually send the response if not already done.
        except TimeoutError as e:
            # a read or a write timed out. Discard this connection
            self.log_error("Request timed out: %r", e)
            self.close_connection = True
            return


def _mock_response() -> mock.AsyncMock:
    """Mock a 200 response."""
    mock_response = mock.AsyncMock(spec=ClientResponse)
    mock_response.status = 200
    mock_response.read.return_value = b""
    mock_response.reason = "OK"
    mock_response.headers = {}
    return mock_response


class MockSession(mock.AsyncMock):
    def __init__(self) -> None:
        super().__init__(spec_set=ClientSession)
        mock_request_ctx = mock.AsyncMock()
        mock_request_ctx.__aenter__.return_value = _mock_response()
        self.request.return_value = mock_request_ctx

    @classmethod
    def klass(cls) -> mock.Mock:
        return mock.Mock(return_value=cls())


class MockConnector(mock.AsyncMock):
    def __init__(self) -> None:
        super().__init__(spec_set=TCPConnector)

    @classmethod
    def klass(cls) -> mock.Mock:
        return mock.Mock(return_value=cls())


def _get_open_port() -> int:
    s = socket.socket(socket.AF_INET, type=socket.SOCK_STREAM)
    s.bind(("127.0.0.1", 0))
    _, port = s.getsockname()
    s.close()
    return port


def _start_server(server: HTTPServer) -> None:
    server.serve_forever()


def _stop_server(server: HTTPServer, thread: threading.Thread) -> None:
    server.shutdown()
    thread.join()


def _create_server(port: int, handler: Callable[[Any], BaseHTTPRequestHandler]) -> Callable[[None], None]:
    server = HTTPServer(("localhost", port), handler)
    server_thread = Thread(target=_start_server, args=(server,))
    server_thread.start()
    return partial(_stop_server, server, server_thread)


# The following test suite takes more time due to starting a local web server. We can probably afford to skip these
# tests while running tests locally.
@long_test
class TestTransportRequest(unittest.IsolatedAsyncioTestCase):
    transport = AioTransport("test-client")
    request_handler = mock.Mock()
    port: int
    url: str

    @classmethod
    def setUpClass(cls) -> None:
        cls.port = _get_open_port()
        cls.url = f"http://localhost:{cls.port}/test"
        cls.addClassCleanup(_create_server(cls.port, partial(EchoHandler, cls.request_handler)))

    def tearDown(self) -> None:
        # Reset the mock object after each test.
        self.request_handler.reset_mock()

    async def asyncSetUp(self) -> None:
        await self.transport.open()

    async def asyncTearDown(self) -> None:
        await self.transport.close()

    def resource_url(self, path: str | None = None) -> ParseResult:
        """Get the full URL for a resource."""
        resolved = self.url.rstrip("/")
        if path is not None:
            resolved += "/" + path.lstrip("/")
        return urlparse(resolved)

    async def request_and_assert(
        self,
        method: RequestMethod,
        relpath: str | None = None,
        headers: Mapping[str, str] | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
        expect_hits: int = 1,
    ) -> HTTPResponse:
        """Send a request and assert that the server handled the correct number of requests.

        :param method: HTTP method.
        :param relpath: Relative path to the resource.
        :param headers: Request headers.
        :param post_params: Request post parameters.
        :param body: Request body.
        :param request_timeout: Request timeout.
        :param expect_hits: Expected number of hits on the request handler.

        :return: The server response.
        """
        uri = self.resource_url(relpath)
        try:
            response = await self.transport.request(
                method=method,
                url=uri.geturl(),
                headers=HTTPHeaderDict(headers or {}),
                post_params=post_params,
                body=body,
                request_timeout=request_timeout,
            )
            self.assertEqual(expect_hits, self.request_handler.call_count)
            self.request_handler.assert_has_calls([mock.call(method, uri.path)] * expect_hits)
        finally:
            if expect_hits == 0:
                self.request_handler.assert_not_called()
        return response

    async def test_post_params_encoded(self) -> None:
        """Test that post parameters are correctly encoded."""
        post_params = [("text1", "abc"), ("text2", "foo bar £".encode("utf-8"))]
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        response = await self.request_and_assert(RequestMethod.POST, headers=headers, post_params=post_params)
        self.assertEqual(b"text1=abc&text2=foo+bar+%C2%A3", response.data)

    async def test_post_params_form_data(self) -> None:
        """Test that post parameters are correctly encoded as form data."""
        post_params = [("text1", "abc"), ("text2", "foo bar £".encode("utf-8"))]
        headers = {"Content-Type": "multipart/form-data"}
        writer = multipart.MultipartWriter("form-data", boundary="foo")
        # Make a reproducible boundary code
        with mock.patch("aiohttp.multipart.MultipartWriter") as MultiPartWriter:
            MultiPartWriter.return_value = writer
            response = await self.request_and_assert(RequestMethod.POST, headers=headers, post_params=post_params)
        expected = (
            b"--foo\r\n"
            b"Content-Type: text/plain; charset=utf-8\r\n"
            b'Content-Disposition: form-data; name="text1"\r\n'
            b"\r\n"
            b"abc\r\n"
            b"--foo\r\n"
            b"Content-Type: application/octet-stream\r\n"
            b'Content-Disposition: form-data; name="text2"; filename="text2"\r\n'
            b"\r\n"
            b"foo bar \xc2\xa3\r\n"
            b"--foo--\r\n"
        )
        self.assertEqual(expected, response.data)

    async def test_post_params_bad_content_type(self) -> None:
        """Test that an error is raised when the content type is not supported."""
        headers = {"Content-Type": "application"}
        with self.assertRaises(TransportError) as ctx:
            await self.request_and_assert(RequestMethod.POST, headers=headers, expect_hits=0)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = (
            "Cannot prepare a request message with the provided arguments. "
            "Please check that your arguments match the declared content type."
        )
        self.assertEqual(expect_msg, str(ctx.exception))

    async def test_post_params_and_body(self) -> None:
        """Test that an error is raised when both post parameters and a body are provided."""
        with self.assertRaises(ClientValueError):
            await self.request_and_assert(
                RequestMethod.GET,
                post_params={},  # type: ignore
                body="body",
                expect_hits=0,
            )

    async def test_head_request(self) -> None:
        """Test sending a HEAD request."""
        await self.request_and_assert(RequestMethod.HEAD)

    async def test_get_request(self) -> None:
        """Test sending a GET request."""
        response = await self.request_and_assert(RequestMethod.GET)
        expected_headers = {"Host": f"localhost:{self.port}", "User-Agent": "test-client"}
        for field, value in expected_headers.items():
            self.assertEqual(value, response.headers[field])

    async def test_no_content_type(self) -> None:
        """Test sending a request without a content type."""
        body = {"Coordinate": {"Latitude": 90, "Longitude": 90}}
        response = await self.request_and_assert(RequestMethod.PUT, body=body)
        self.assertEqual(b'{"Coordinate": {"Latitude": 90, "Longitude": 90}}', response.data)

    async def test_json_content_type(self) -> None:
        """Test sending a request with a JSON body."""
        headers = {"Content-Type": "application/json"}
        body = {"Coordinate": {"Latitude": 90, "Longitude": 90}}
        response = await self.request_and_assert(RequestMethod.PUT, body=body, headers=headers)
        self.assertEqual(b'{"Coordinate": {"Latitude": 90, "Longitude": 90}}', response.data)

    async def test_str_body(self) -> None:
        """Test sending a request with a string body."""
        body = "Header1,Header2\nText,1.2"
        response = await self.request_and_assert(RequestMethod.PUT, body=body)
        self.assertEqual(body.encode("utf-8"), response.data)

    async def test_bytes_body(self) -> None:
        """Test sending a request with a bytes body."""
        body = "foo bar £".encode("utf-8")
        response = await self.request_and_assert(RequestMethod.PUT, body=body)
        self.assertEqual(body, response.data)

    @parameterized.expand(
        [
            (1, None),
            (1.5, None),
            (4, 1),
            (1, 3),
        ]
    )
    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    async def test_timeout(
        self, connect_timeout: float, read_timeout: float | None, mock_session_klass: mock.Mock
    ) -> None:
        """Test setting a timeout on the request.

        This test has consistently caused problems testing against a real server due to varying response times. The
        underlying aiohttp.ClientSession object is now mocked to avoid these issues, and we are to assume that the
        timeout implementation in aiohttp is as advertised.
        """
        # Use a transport object that only allows one attempt.
        transport = AioTransport("test-client", max_attempts=1)
        await transport.open()

        # Mock the session object to raise a timeout error.
        mock_session: MockSession = mock_session_klass.return_value
        mock_session.request.side_effect = TimeoutError()

        # Send the request.
        request_timeout = connect_timeout if read_timeout is None else (connect_timeout, read_timeout)
        with self.assertRaises(TransportError) as ctx:
            await transport.request(
                RequestMethod.GET,
                self.url,
                request_timeout=request_timeout,
            )

        # Check the exception.
        cause = ctx.exception.caused_by
        self.assertIsInstance(cause, RetryError)
        (cause,) = cause.exceptions
        self.assertIsInstance(cause, TimeoutError)

        # Check the request call.
        expect_timeout = (
            ClientTimeout(total=connect_timeout)
            if read_timeout is None
            else ClientTimeout(sock_connect=connect_timeout, sock_read=read_timeout)
        )
        mock_session.request.assert_called_once_with(
            allow_redirects=False,
            method="GET",
            url=self.url,
            headers=HTTPHeaderDict({"User-Agent": "test-client"}),
            data=None,
            timeout=expect_timeout,
            proxy=None,
        )

    async def test_retries(self) -> None:
        """Test that the transport retries the request."""
        # Use a transport object that allows 5 attempts with no delay.
        await self.transport.close()
        self.transport = AioTransport("test-client", max_attempts=5, backoff_method=BackoffLinear(0))
        await self.transport.open()

        # Test that the transport retries the request.
        with self.assertRaises(TransportError) as ctx:
            await self.request_and_assert(RequestMethod.GET, relpath="/retry", expect_hits=5)

        # Check the exception.
        cause = ctx.exception.caused_by
        self.assertIsInstance(cause, RetryError)
        self.assertEqual(5, len(cause.exceptions))

    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    @mock.patch("aiohttp.TCPConnector", new_callable=MockConnector.klass)
    async def test_open(self, mock_connector_klass: mock.Mock, mock_session_klass: mock.Mock) -> None:
        """Test opening the transport."""
        mock_connector: MockConnector = mock_connector_klass.return_value
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session and connector.
        transport = AioTransport("test-client", max_attempts=4)

        # Check that the transport is not opened yet.
        mock_connector_klass.assert_not_called()
        mock_session_klass.assert_not_called()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

        # Open the transport and check that the session and connector are created.
        await transport.open()
        mock_connector_klass.assert_called_once_with(ssl=None, limit=4)
        mock_session_klass.assert_called_once_with(
            connector=mock_connector, skip_auto_headers=["Accept", "Accept-Encoding"]
        )
        await transport.request(RequestMethod.GET, self.url)
        mock_session.request.assert_called_once()

    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    @mock.patch("aiohttp.TCPConnector", new_callable=MockConnector.klass)
    async def test_aenter(self, mock_connector_klass: mock.Mock, mock_session_klass: mock.Mock) -> None:
        """Test entering the transport context."""
        mock_connector: MockConnector = mock_connector_klass.return_value
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session and connector.
        transport = AioTransport("test-client", max_attempts=4)

        # Check that the transport is not opened yet.
        mock_connector_klass.assert_not_called()
        mock_session_klass.assert_not_called()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

        # Open the transport and check that the session and connector are created.
        async with transport:
            mock_connector_klass.assert_called_once_with(ssl=None, limit=4)
            mock_session_klass.assert_called_once_with(
                connector=mock_connector, skip_auto_headers=["Accept", "Accept-Encoding"]
            )
            await transport.request(RequestMethod.GET, self.url)
            mock_session.request.assert_called_once()

    @mock.patch("asyncio.sleep")
    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    async def test_close(self, mock_session_klass: mock.Mock, mock_sleep: mock.Mock) -> None:
        """Test closing the transport."""
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session.
        transport = AioTransport("test-client", close_grace_period_ms=20)
        # Open the transport.
        await transport.open()
        await transport.request(RequestMethod.GET, self.url)
        mock_session.request.assert_called_once()

        # Check that the session is not closed yet.
        mock_session.close.assert_not_called()

        # Close the transport and check that the session is closed.
        await transport.close()
        mock_session.close.assert_called_once()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

        # Check that the correct duration of sleep was called during the close process.
        mock_sleep.assert_called_once_with(20 / 1000)

    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    async def test_aexit(self, mock_session_klass: mock.Mock) -> None:
        """Test exiting the transport context."""
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session.
        transport = AioTransport("test-client")
        # Open the transport.
        async with transport:
            await transport.request(RequestMethod.GET, self.url)
            mock_session.request.assert_called_once()

            # Check that the session is not closed yet.
            mock_session.close.assert_not_called()

        # Exiting the context should close the session.
        mock_session.close.assert_called_once()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    @mock.patch("aiohttp.TCPConnector", new_callable=MockConnector.klass)
    async def test_reentrant(self, mock_connector_klass: mock.Mock, mock_session_klass: mock.Mock) -> None:
        """Test opening and closing the transport multiple times."""
        mock_connector: MockConnector = mock_connector_klass.return_value
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session and connector.
        transport = AioTransport("test-client", max_attempts=4)

        # Check that the transport is not opened yet.
        mock_connector_klass.assert_not_called()
        mock_session_klass.assert_not_called()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

        # Open the transport and check that the session and connector are created.
        await transport.open()
        mock_connector_klass.assert_called_once_with(ssl=None, limit=4)
        mock_session_klass.assert_called_once_with(
            connector=mock_connector, skip_auto_headers=["Accept", "Accept-Encoding"]
        )
        await transport.request(RequestMethod.GET, self.url)
        mock_session.request.assert_called_once()
        mock_session.close.assert_not_called()

        mock_session.request.reset_mock()  # Reset the call count.

        # Open the transport again and check that the session and connector are not created again.
        await transport.open()
        mock_connector_klass.assert_called_once_with(ssl=None, limit=4)
        mock_session_klass.assert_called_once_with(
            connector=mock_connector, skip_auto_headers=["Accept", "Accept-Encoding"]
        )
        await transport.request(RequestMethod.GET, self.url)
        mock_session.request.assert_called_once()
        mock_session.close.assert_not_called()

        mock_session.request.reset_mock()  # Reset the call count.

        # Close the transport and check that the session is not closed.
        await transport.close()
        mock_session.close.assert_not_called()
        await transport.request(RequestMethod.GET, self.url)
        mock_session.request.assert_called_once()

        # Close the transport again and check that the session is closed.
        await transport.close()
        mock_session.close.assert_called_once()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    @mock.patch("aiohttp.TCPConnector", new_callable=MockConnector.klass)
    async def test_reentrant_ctx_manager(self, mock_connector_klass: mock.Mock, mock_session_klass: mock.Mock) -> None:
        """Test context manager reentrancy."""
        mock_connector: MockConnector = mock_connector_klass.return_value
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session and connector.
        transport = AioTransport("test-client", max_attempts=4)

        # Check that the transport is not opened yet.
        mock_connector_klass.assert_not_called()
        mock_session_klass.assert_not_called()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

        # Open the transport and check that the session and connector are created.
        async with transport:
            mock_connector_klass.assert_called_once_with(ssl=None, limit=4)
            mock_session_klass.assert_called_once_with(
                connector=mock_connector, skip_auto_headers=["Accept", "Accept-Encoding"]
            )
            await transport.request(RequestMethod.GET, self.url)
            mock_session.request.assert_called_once()
            mock_session.close.assert_not_called()

            mock_session.request.reset_mock()  # Reset the call count.

            # Open the transport again and check that the session and connector are not created again.
            async with transport:
                mock_connector_klass.assert_called_once_with(ssl=None, limit=4)
                mock_session_klass.assert_called_once_with(
                    connector=mock_connector, skip_auto_headers=["Accept", "Accept-Encoding"]
                )
                await transport.request(RequestMethod.GET, self.url)
                mock_session.request.assert_called_once()
                mock_session.close.assert_not_called()

                mock_session.request.reset_mock()  # Reset the call count.

            # Close the transport and check that the session is not closed.
            mock_session.close.assert_not_called()
            await transport.request(RequestMethod.GET, self.url)
            mock_session.request.assert_called_once()

        # Close the transport again and check that the session is closed.
        mock_session.close.assert_called_once()
        with self.assertRaises(TransportError) as ctx:
            await transport.request(RequestMethod.GET, self.url)
        self.assertIsNone(ctx.exception.caused_by)
        expect_msg = "Cannot make a request before the transport has been opened, or after it has been closed."
        self.assertEqual(expect_msg, str(ctx.exception))

    @mock.patch("aiohttp.ClientSession", new_callable=MockSession.klass)
    async def test_proxy(self, mock_session_klass: mock.Mock) -> None:
        """Test creating AioTransport with a proxy uses the proxy in the request."""
        mock_session: MockSession = mock_session_klass.return_value

        # Create a new transport object that picks up the mocked session.
        transport = AioTransport("test-client", proxy="https://example.com:8080")

        # Open the transport and check that the session is created with the proxy.
        await transport.open()
        await transport.request(RequestMethod.GET, self.url)
        mock_session.request.assert_called_once_with(
            allow_redirects=False,
            method="GET",
            url=self.url,
            headers=HTTPHeaderDict({"User-Agent": "test-client"}),
            data=None,
            timeout=None,
            proxy="https://example.com:8080",
        )

    async def test_bg_task_cancelled_during_close(self):
        """Test for cancelling background tasks mid close."""

        transport = AioTransport(user_agent="test-agent")
        await transport.open()

        original_close_method = transport._AioTransport__context.close
        close_finished_event = asyncio.Event()

        async def context_close_wrapper():
            """Wait indefinitely at the end of the coroutine to allow for cancellation."""
            await original_close_method()
            close_finished_event.set()
            await asyncio.Event().wait()

        with patch.object(transport._AioTransport__context, "close", new=context_close_wrapper):
            # Schedule the close coroutine with a wait after the context is closed
            close_task = asyncio.create_task(transport.close())

            # Wait for the original close coroutine to finish
            await close_finished_event.wait()

            # Cancel the future, this should happen before the context_close_wrapper is able to exit
            close_task.cancel()

            # Context should be cleaned up even though we cancelled
            self.assertIsNone(transport._AioTransport__context)

            # Open the transport again
            await transport.open()

            # Make a request
            await transport.request(RequestMethod.GET, "https://example.com", HTTPHeaderDict())

            # Check that the context was recreated
            self.assertIsNotNone(transport._AioTransport__context)

            # Close the transport
            await transport.close()

            # Check that the context was closed
            self.assertIsNone(transport._AioTransport__context)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from datetime import datetime, timezone
from uuid import UUID

from evo import logging
from evo.common import (
    APIConnector,
    BaseAPIClient,
    Environment,
    HealthCheckType,
    ServiceHealth,
    ServiceUser,
)
from evo.common.utils import get_service_health

from .data import (
    Association,
    AssociationMetadata,
    CategoryColormap,
    ColormapMetadata,
    ContinuousColormap,
    DiscreteColormap,
)
from .endpoints.api import AssociationsApi, ColormapsApi
from .endpoints.models import (
    AttributeAssociationData,
    AttributeAssociationsData,
    AttributeColormapAssociation,
    CategoryColormapData,
    CategoryColormapResponse,
    ColormapData,
    ColormapTypeEnum,
    ContinuousColormapData,
    ContinuousColormapResponse,
    DiscreteColormapData,
    DiscreteColormapResponse,
)
from .exceptions import UnknownColormapResponse, UnknownColormapType

logger = logging.getLogger("colormap.client")

__all__ = [
    "ColormapAPIClient",
]


def _colors_from_endpoint_model(
    model: ContinuousColormapResponse | DiscreteColormapResponse | CategoryColormapResponse,
) -> list[list[int]]:
    """Get colors list from an endpoint model."""
    return [[channel.root for channel in color.root] for color in model.colors]


def _with_timezone(dt: datetime) -> datetime:
    """Add timestamps to naive datetimes returned from the API."""
    if dt.tzinfo is not None:  # future proofing
        return dt
    return dt.replace(tzinfo=timezone.utc)


class ColormapAPIClient(BaseAPIClient):
    def __init__(self, environment: Environment, connector: APIConnector) -> None:
        """
        :param environment: The environment object
        :param connector: The connector object.
        """
        super().__init__(environment, connector)
        self._colourmaps_api = ColormapsApi(connector=connector)
        self._associations_api = AssociationsApi(connector=connector)

    def _colormap_metadata_from_endpoint_model(
        self, model: ContinuousColormapResponse | DiscreteColormapResponse | CategoryColormapResponse
    ) -> ColormapMetadata:
        """Creates a ColormapMetadata object from an endpoint model.

        :param model: Colormap response model from the endpoint.

        :return: A ColormapMetadata object.

        :raises UnknownColormapResponse: If the response model is not recognized.
        """
        match model:
            case ContinuousColormapResponse():
                colormap = ContinuousColormap(
                    colors=_colors_from_endpoint_model(model),
                    attribute_controls=model.attribute_controls,
                    gradient_controls=model.gradient_controls,
                )
            case DiscreteColormapResponse():
                colormap = DiscreteColormap(
                    colors=_colors_from_endpoint_model(model),
                    end_inclusive=model.end_inclusive,
                    end_points=model.end_points,
                )
            case CategoryColormapResponse():
                colormap = CategoryColormap(
                    colors=_colors_from_endpoint_model(model),
                    map=model.map,
                )
            case _:
                raise UnknownColormapResponse(f"Unexpected colormap response type: {type(model)}")

        return ColormapMetadata(
            environment=self._environment,
            id=model.id,
            name=model.name,
            created_at=_with_timezone(model.created_at),
            created_by=ServiceUser(id=model.created_by, name=None, email=None),
            modified_at=_with_timezone(model.modified_at),
            modified_by=ServiceUser(id=model.modified_by, name=None, email=None),
            colormap=colormap,
            self_link=str(model.self_link),
        )

    def _association_metadata_from_endpoint_model(
        self, model: AttributeColormapAssociation, object_id: UUID
    ) -> AssociationMetadata:
        """Creates an AssociationMetadata object from an endpoint model.

        :param model: Association response model from the endpoint.

        :return: An AssociationMetadata object.
        """
        return AssociationMetadata(
            environment=self._environment,
            id=model.id,
            created_at=_with_timezone(model.created_at),
            created_by=ServiceUser(id=model.created_by, name=None, email=None),
            modified_at=_with_timezone(model.modified_at),
            modified_by=ServiceUser(id=model.modified_by, name=None, email=None),
            attribute_id=model.attribute_id,
            colormap_id=model.colormap_id,
            object_id=object_id,
            colormap_uri=str(model.colormap_uri),
            self_link=str(model.self_link),
            name=f"Association between {model.attribute_id} and {model.colormap_id}",
        )

    async def get_service_health(self, check_type: HealthCheckType = HealthCheckType.FULL) -> ServiceHealth:
        """Get the health of the service.

        :param check_type: The type of health check to perform.

        :return: A ServiceHealth object.

        :raises EvoApiException: If the API returns an unexpected status code.
        :raises ClientValueError: If the response is not a valid service health check response.
        """
        return await get_service_health(self._connector, "visualization", check_type=check_type)

    async def create_colormap(
        self,
        colormap: ContinuousColormap | DiscreteColormap | CategoryColormap,
        name: str,
    ) -> ColormapMetadata:
        """Create a new colormap.

        :param colormap: The colormap data.
        :param name: The name of the colormap.

        :return: The created colormap.

        :raises UnknownColormapType: If the colormap type is not recognized.
        :raises EvoApiException: If the API returns an unexpected status code.
        """
        match colormap:
            case ContinuousColormap():
                colormap_data = ContinuousColormapData.model_validate(
                    {
                        "colors": colormap.colors,
                        "attribute_controls": colormap.attribute_controls,
                        "gradient_controls": colormap.gradient_controls,
                        "name": name,
                        "dtype": ColormapTypeEnum.continuous,
                    }
                )
            case DiscreteColormap():
                colormap_data = DiscreteColormapData.model_validate(
                    {
                        "colors": colormap.colors,
                        "end_inclusive": colormap.end_inclusive,
                        "end_points": colormap.end_points,
                        "name": name,
                        "dtype": ColormapTypeEnum.discrete,
                    }
                )
            case CategoryColormap():
                colormap_data = CategoryColormapData.model_validate(
                    {
                        "colors": colormap.colors,
                        "map": colormap.map,
                        "name": name,
                        "dtype": ColormapTypeEnum.category,
                    }
                )
            case _:
                raise UnknownColormapType(f"Unexpected colormap type: {type(colormap)}")

        response = await self._colourmaps_api.post_colormap(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            colormap_data=ColormapData.model_validate(colormap_data),
        )
        return self._colormap_metadata_from_endpoint_model(response.root)

    async def get_colormap_by_id(self, colormap_id: UUID) -> ColormapMetadata:
        """Get a colormap by ID.

        :param colormap_id: The UUID of the colormap.

        :return: A ColormapMetadata object representation of the colormap.

        :raises EvoApiException: If the API returns an unexpected status code.
        """
        response = await self._colourmaps_api.get_colormap_by_id(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            colormap_id=str(colormap_id),
        )
        return self._colormap_metadata_from_endpoint_model(response.root)

    async def get_colormap_collection(self) -> list[ColormapMetadata]:
        """Get all the colormaps in the current workspace.

        :return: A list of ColormapMetadata object representations of the colormaps.

        :raises EvoApiException: If the API returns an unexpected status code.
        """
        response = await self._colourmaps_api.get_colormap_collection(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
        )
        return [self._colormap_metadata_from_endpoint_model(model) for model in response.colormaps]

    async def create_association(self, object_id: UUID, association: Association) -> AssociationMetadata:
        """Associate an existing colormap with a geoscience object.

        :param object_id: The UUID of the geoscience object.
        :param association: The association data

        :return: An AssociationMetadata object representation of the association.

        :raises EvoApiException: If the API returns an unexpected status code.
        """
        response = await self._associations_api.post_association(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            attribute_association_data=AttributeAssociationData.model_validate(
                {"attribute_id": association.attribute_id, "colormap_id": str(association.colormap_id)}
            ),
            object_id=str(object_id),
        )
        return self._association_metadata_from_endpoint_model(response, object_id)

    async def create_batch_associations(
        self,
        object_id: UUID,
        associations: list[Association],
    ) -> list[AssociationMetadata]:
        """Create multiple associations for a colormap to multiple object attributes.

        :param object_id: The UUID of the geoscience object.
        :param associations: A list of dictionaries containing attribute_id and colormap_id pairs.

        :return: A list of AssociationMetadata object representations of the associations.

        :raises EvoApiException: If the API returns an unexpected status code.
        """
        if len(associations) == 0:
            raise ValueError("Associations list must not be empty.")

        def _chunk_list(lst):
            """Yield successive chunks from list."""
            chunk_size = 128
            for i in range(0, len(lst), chunk_size):
                yield lst[i : i + chunk_size]

        all_associations_metadata = []
        for chunk in _chunk_list(associations):
            attribute_associations_data = AttributeAssociationsData.model_validate(
                {
                    "associations": [
                        AttributeAssociationData.model_validate(
                            {"attribute_id": association.attribute_id, "colormap_id": str(association.colormap_id)}
                        )
                        for association in chunk
                    ]
                }
            )

            response = await self._associations_api.post_batch_associations(
                org_id=str(self._environment.org_id),
                workspace_id=str(self._environment.workspace_id),
                object_id=str(object_id),
                attribute_associations_data=attribute_associations_data,
            )
            all_associations_metadata.extend(
                self._association_metadata_from_endpoint_model(model, object_id) for model in response.associations
            )

        return all_associations_metadata

    async def get_association(self, object_id: UUID, association_id: UUID) -> AssociationMetadata:
        """Get the specific colormap association for a geoscience object given an association UUID.

        :param object_id: The UUID of the geoscience object.
        :param association_id: The UUID of the association.

        :return: An AssociationMetadata object representation of the association.

        :raises EvoApiException: If the API returns an unexpected status code.
        """
        response = await self._associations_api.get_association_by_id(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            object_id=str(object_id),
            association_id=str(association_id),
        )
        return self._association_metadata_from_endpoint_model(response, object_id)

    async def get_association_collection(self, object_id: UUID) -> list[AssociationMetadata]:
        """Get all associations for a geoscience object.

        :param object_id: The UUID of the geoscience object.

        :return: A list of AssociationMetadata object representations of the associations.

        :raises EvoApiException: If the API returns an unexpected status code.
        """
        response = await self._associations_api.get_association_collection(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            object_id=str(object_id),
        )
        return [self._association_metadata_from_endpoint_model(model, object_id) for model in response.associations]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pyarrow as pa

from evo import logging

from ..exceptions import TableFormatError
from ._types import Table
from .tables import ArrowTableFormat, KnownTableFormat

__all__ = [
    "all_known_formats",
    "get_known_format",
]

logger = logging.getLogger("object.table_formats")

_float_array_formats: list[KnownTableFormat] = [
    KnownTableFormat(
        name="float-array-1", columns=[pa.float64()], field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="float-array-2", columns=[pa.float64()] * 2, field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="float-array-3", columns=[pa.float64()] * 3, field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="float-array-6", columns=[pa.float64()] * 6, field_names=["data", "length", "width", "data_type"]
    ),
    # float-array-md MUST be last so that formats with fixed sizes are matched first.
    KnownTableFormat(
        name="float-array-md", columns=[pa.float64(), ...], field_names=["data", "length", "width", "data_type"]
    ),
]


_index_array_formats: list[KnownTableFormat] = [
    KnownTableFormat(name="index-array-1", columns=[pa.uint64()], field_names=["data", "length", "width", "data_type"]),
    KnownTableFormat(
        name="index-array-2", columns=[pa.uint64()] * 2, field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="index-array-3", columns=[pa.uint64()] * 3, field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="index-array-4", columns=[pa.uint64()] * 4, field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="index-array-8", columns=[pa.uint64()] * 8, field_names=["data", "length", "width", "data_type"]
    ),
]


_integer_array_formats: list[KnownTableFormat] = [
    KnownTableFormat(
        name="integer-array-1-int32", columns=[pa.int32()], field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="integer-array-1-int64", columns=[pa.int64()], field_names=["data", "length", "width", "data_type"]
    ),
    # integer-array-md-* MUST be last so that formats with fixed sizes are matched first.
    KnownTableFormat(
        name="integer-array-md-int32", columns=[pa.int32(), ...], field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(
        name="integer-array-md-int64", columns=[pa.int64(), ...], field_names=["data", "length", "width", "data_type"]
    ),
]


_boolean_array_formats: list[KnownTableFormat] = [
    KnownTableFormat(name="bool-array-1", columns=[pa.bool_()], field_names=["data", "length", "data_type"]),
    # bool-array-md MUST be last so that formats with fixed sizes are matched first.
    KnownTableFormat(
        name="bool-array-md", columns=[pa.bool_(), ...], field_names=["data", "length", "width", "data_type"]
    ),
]


_other_array_formats: list[KnownTableFormat] = [
    # Simple arrays.
    KnownTableFormat(
        name="brep-container-brep", columns=[pa.uint8()], field_names=["data", "length", "width", "data_type"]
    ),
    KnownTableFormat(name="color-array", columns=[pa.uint32()], field_names=["data", "length", "data_type"]),
    KnownTableFormat(
        name="date-time-array", columns=[pa.timestamp("us", tz="UTC")], field_names=["data", "length", "data_type"]
    ),
    KnownTableFormat(name="string-array", columns=[pa.string()], field_names=["data", "length", "data_type"]),
    # Complex arrays.
    KnownTableFormat(
        name="downhole-collection-location-holes",
        columns=[pa.int32(), pa.uint64(), pa.uint64()],
        field_names=["data", "length", "width", "data_type"],
    ),
    KnownTableFormat(
        name="unstructured-grid-geometry-cells",
        columns=[pa.int32(), pa.uint64(), pa.int32()],
        field_names=["data", "length", "width", "data_type"],
    ),
    KnownTableFormat(
        name="lines-2d-indices",
        columns=[pa.uint64(), pa.uint64(), pa.float64()],
        field_names=["data", "length", "width", "data_type"],
    ),
    KnownTableFormat(
        name="lookup-table-int32",
        columns=[pa.int32(), pa.string()],
        field_names=["data", "length", "keys_data_type", "values_data_type"],
    ),
    KnownTableFormat(
        name="lookup-table-int64",
        columns=[pa.int64(), pa.string()],
        field_names=["data", "length", "keys_data_type", "values_data_type"],
    ),
]


all_known_formats: list[KnownTableFormat] = [
    *_float_array_formats,
    *_index_array_formats,
    *_integer_array_formats,
    *_boolean_array_formats,
    *_other_array_formats,
]


def get_known_format(table: Table) -> KnownTableFormat:
    """Get the known table format that best matches the provided table.

    If both a multidimensional format and a format with fixed dimensions would match, the format with fixed dimensions
    will be returned.

    :param table: The actual table to match.

    :return: The known format that best matches the provided table.

    :raises TableFormatError: If the provided table does not match a known format.
    """
    actual_format = ArrowTableFormat.from_schema(table.schema)
    logger.debug(f"Matching format for {actual_format.name}")
    for known_format in all_known_formats:
        if known_format.is_provided_by(actual_format):
            logger.debug(f"{known_format.name} is provided by {actual_format.name}")
            return known_format

    msg = f"Could not resolve Geoscience Object Schema for {actual_format.name}"
    logger.error(msg)
    raise TableFormatError(msg)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from uuid import UUID

from data import load_test_data
from evo.common import ServiceUser
from evo.common.test_tools import TestWithConnector, TestWithDownloadHandler, TestWithUploadHandler, utc_datetime
from evo.files import FileAPIDownload, FileAPIUpload, FileMetadata

TEST_DATA = """
x,y,z
10689.54607,100813.2874,442.172449
10690.87819,100812.6204,437.399555
10692.21062,100811.9502,432.627182
10693.54337,100811.277,427.85533
10694.87638,100810.6009,423.083966
10696.06353,100809.9976,418.831067
10725.47983,100777.2912,448.471682
10727.51038,100775.9784,444.022761
10729.54666,100774.6657,439.576449
10731.58199,100773.3566,435.128615
""".lstrip().encode("utf-8")

FILE_ID = UUID(int=5)
VERSION_ID = "123456"
INITIAL_URL = "https://unit.test/initial/url"


class TestFileAPIDownload(TestWithConnector, TestWithDownloadHandler):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithDownloadHandler.setUp(self)
        self.setup_download_handler(TEST_DATA)
        self.metadata = FileMetadata(
            environment=self.environment,
            id=FILE_ID,
            name="file.csv",
            created_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            created_at=utc_datetime(2020, 1, 1, 1, 30),
            modified_by=ServiceUser(
                id=UUID(int=16),
                name="x y",
                email="test@example.com",
            ),
            modified_at=utc_datetime(2020, 1, 1, 1, 30),
            parent="/some/path",
            version_id=VERSION_ID,
            size=len(TEST_DATA),
        )
        self.download = FileAPIDownload(connector=self.connector, metadata=self.metadata, initial_url=INITIAL_URL)

    def test_label(self) -> None:
        expected = f"{self.metadata.id}?version_id={self.metadata.version_id}"
        self.assertEqual(expected, self.download.label)

    def test_metadata(self) -> None:
        self.assertEqual(self.metadata, self.download.metadata)

    def test_get_cache_location(self) -> None:
        expected = self.cache.get_location(self.environment, "filev2") / str(FILE_ID) / VERSION_ID / self.metadata.name
        self.assertEqual(expected, self.download._get_cache_location(self.cache))

    async def test_get_download_url(self) -> None:
        # Test the initial URL is used first.
        first = await self.download.get_download_url()
        self.assertEqual(INITIAL_URL, first)

        # No requests should be made when using the initial URL.
        self.transport.assert_no_requests()

        # Test that a new URL is generated when the initial URL is used up.
        get_file_response = load_test_data("get_file.json")
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_file_response), headers={"Content-Type": "application/json"}
        ):
            second = await self.download.get_download_url()
        self.assertEqual(get_file_response["download"], second)

    async def test_download_to_path(self) -> None:
        dest = self.cache.root / "test_download_to_path.csv"
        await self.download.download_to_path(dest, self.transport)

        # Check the file was downloaded correctly.
        self.assertEqual(TEST_DATA, dest.read_bytes())
        self.assert_download_requests(INITIAL_URL)

    async def test_download_to_cache(self) -> None:
        dest = await self.download.download_to_cache(self.cache, self.transport)

        # Check the download location is correct.
        expected = self.cache.get_location(self.environment, "filev2") / str(FILE_ID) / VERSION_ID / self.metadata.name
        self.assertEqual(expected, dest)

        # Check the file was downloaded correctly.
        self.assertEqual(TEST_DATA, dest.read_bytes())
        self.assert_download_requests(INITIAL_URL)


class TestFileAPIUpload(TestWithConnector, TestWithUploadHandler):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithUploadHandler.setUp(self)
        self.upload = FileAPIUpload(
            connector=self.connector,
            environment=self.environment,
            file_id=FILE_ID,
            version_id=VERSION_ID,
            initial_url=INITIAL_URL,
        )

    def test_label(self) -> None:
        expected = f"{FILE_ID}?version_id={VERSION_ID}"
        self.assertEqual(expected, self.upload.label)

    async def test_get_upload_url(self) -> None:
        # Test the initial URL is used first.
        first = await self.upload.get_upload_url()
        self.assertEqual(INITIAL_URL, first)

        # No requests should be made when using the initial URL.
        self.transport.assert_no_requests()

        # Test that a new URL is generated when the initial URL is used up.
        update_file_response = load_test_data("update_file.json")
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(update_file_response), headers={"Content-Type": "application/json"}
        ):
            second = await self.upload.get_upload_url()
        self.assertEqual(update_file_response["upload"], second)

    async def test_upload_from_path(self) -> None:
        source = self.cache.root / "test_upload_from_path.csv"
        source.write_bytes(TEST_DATA)
        await self.upload.upload_from_path(source, self.transport)

        # Check the file was uploaded correctly.
        uploaded = await self.handler.get_committed()
        self.assertEqual(TEST_DATA, uploaded)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from unittest import mock
from urllib.parse import parse_qs, urlparse

from parameterized import parameterized

from evo.common.test_tools import long_test
from evo.oauth import EvoScopes, OAuthError, OAuthRedirectHandler

from ._helpers import (
    AUTHORIZATION_CODE,
    REDIRECT_URL,
    STATE_TOKEN,
    VERIFIER_TOKEN,
    TestWithMockOAuthConnector,
    TestWithOAuthConnector,
    get_access_token,
    get_redirect,
    patch_urlsafe_tokens,
    patch_webbrowser_open,
)


class TestOAuthRedirectHandlerWithMockOAuth(TestWithMockOAuthConnector):
    def setUp(self) -> None:
        super().setUp()
        self.handler = OAuthRedirectHandler(oauth_connector=self.connector, redirect_url=REDIRECT_URL)
        self.handler.get_token = mock.AsyncMock(return_value=get_access_token())

    async def authenticate(self) -> None:
        """Authenticate and return the access token."""
        async with self.handler:
            response_text = await get_redirect(state=STATE_TOKEN, code=AUTHORIZATION_CODE)

        self.handler.get_token.assert_called_once_with(STATE_TOKEN, AUTHORIZATION_CODE)
        self.assertEqual(OAuthRedirectHandler._REDIRECT_HTML.decode("utf-8"), response_text)

    async def test_pending_with_success(self) -> None:
        """Test that the handler is pending until authentication is successful"""
        self.assertTrue(self.handler.pending)
        await self.authenticate()
        self.assertFalse(self.handler.pending)

    async def test_pending_with_failed(self) -> None:
        """Test that the handler is pending until authentication fails"""
        self.assertTrue(self.handler.pending)

        class SomeError(Exception): ...

        self.handler.get_token.side_effect = SomeError("Test error")
        await self.authenticate()
        self.assertFalse(self.handler.pending)

        with self.assertRaises(OAuthError):
            await self.handler.get_result()

    async def test_get_result(self) -> None:
        """Test that the handler returns the access token"""
        with self.assertRaises(OAuthError):
            await self.handler.get_result()

        async with self.handler:
            auth = asyncio.create_task(self.handler.get_result())
            redirect = get_redirect()
            self.assertFalse(auth.done())
            result, _ = await asyncio.gather(auth, redirect)
            self.assertTrue(auth.done())

        self.handler.get_token.assert_called_once_with(STATE_TOKEN, AUTHORIZATION_CODE)
        self.assertEqual(get_access_token(), result)

    @parameterized.expand(
        [
            ("no delay", 0),
            ("short delay", 1),
            ("longer delay", 5),
        ]
    )
    @long_test
    async def test_get_result_timeout(self, _label: str, delay: int) -> None:
        """Test that the handler times out waiting for the result."""
        async with self.handler:
            auth = asyncio.create_task(self.handler.get_result(delay))
            self.assertFalse(auth.done())
            with self.assertRaises(OAuthError):
                await auth
            self.assertTrue(auth.done())

    async def test_reentering_context(self) -> None:
        """Test that re-entering the context raises an error."""
        async with self.handler:
            with self.assertRaises(OAuthError):
                async with self.handler:
                    pass

    async def test_exiting_unentered_context(self) -> None:
        """Test that exiting an unentered context raises an error."""
        with self.assertRaises(OAuthError):
            await self.handler.__aexit__(None, None, None)

    async def test_login(self) -> None:
        """Test logging in with the handler."""
        expect_auth_url = "http://unit.test"
        self.handler.create_authorization_url = mock.Mock(return_value=expect_auth_url)

        async with self.handler:
            with patch_webbrowser_open(authenticate=True) as mock_webbrowser_open:
                result = await self.handler.login(EvoScopes.all_evo)

        mock_webbrowser_open.assert_called_once_with(expect_auth_url)
        self.handler.get_token.assert_called_once_with(STATE_TOKEN, AUTHORIZATION_CODE)
        self.assertEqual(get_access_token(), result)

    @parameterized.expand(
        [
            ("no delay", 0),
            ("short delay", 1),
            ("longer delay", 5),
        ]
    )
    @long_test
    async def test_login_timeout(self, _label: str, delay: int) -> None:
        """Test that the handler times out waiting for the user to authenticate."""
        expect_auth_url = "http://unit.test"
        self.handler.create_authorization_url = mock.Mock(return_value=expect_auth_url)

        async with self.handler:
            with patch_webbrowser_open(authenticate=False) as mock_webbrowser_open:
                auth = asyncio.create_task(self.handler.login(EvoScopes.all_evo, delay))
                self.assertFalse(auth.done())
                with self.assertRaises(OAuthError):
                    await auth
                self.assertTrue(auth.done())

        mock_webbrowser_open.assert_called_once_with(expect_auth_url)
        self.handler.get_token.assert_not_called()

    def test_create_authorization_url(self) -> None:
        """Test creating the authorization URL"""
        expected_url = self.get_expected_auth_url(state=STATE_TOKEN, verifier=VERIFIER_TOKEN, scopes=EvoScopes.default)
        with patch_urlsafe_tokens(state=STATE_TOKEN, verifier=VERIFIER_TOKEN):
            actual_url = self.handler.create_authorization_url(EvoScopes.default)

        # Check the auth URL and query parameters.
        self.assertEqual(expected_url, actual_url)

    def test_create_authorization_url_generates_unique_tokens(self) -> None:
        """Test creating the authorization URL generates unique state and verifier tokens."""

        uri_one = self.handler.create_authorization_url(EvoScopes.default)
        query_one = parse_qs(urlparse(uri_one).query)

        uri_two = self.handler.create_authorization_url(EvoScopes.default)
        query_two = parse_qs(urlparse(uri_two).query)

        self.assertNotEqual(query_one["state"], query_two["state"])
        self.assertNotEqual(query_one["code_challenge"], query_two["code_challenge"])


class TestOAuthRedirectHandler(TestWithOAuthConnector):
    def setUp(self) -> None:
        super().setUp()
        self.handler = OAuthRedirectHandler(oauth_connector=self.connector, redirect_url=REDIRECT_URL)

    async def test_get_token_fails_without_state(self) -> None:
        """Test getting an access token fails if the state token has not been generated"""
        with self.assertRaises(OAuthError):
            await self.handler.get_token(AUTHORIZATION_CODE, VERIFIER_TOKEN)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import contextlib
import logging
import unittest
from io import BytesIO
from threading import Lock
from unittest import mock

from parameterized import parameterized

from evo.common.exceptions import RetryError
from evo.common.io import ChunkedIOManager, ChunkedIOTracker, ChunkMetadata
from evo.common.io.exceptions import ChunkedIOError
from evo.common.io.interfaces import IDestination, ISource
from evo.common.utils import BackoffLinear, Retry

logger = logging.getLogger(__name__)


class TestChunkedIOTracker(unittest.TestCase):
    FILE_SIZE = 15
    CHUNK_SIZE = 4
    LAST_CHUNK_ID = 3
    LAST_CHUNK_SIZE = 3

    def setUp(self) -> None:
        self.tracker = ChunkedIOTracker(self.FILE_SIZE, self.CHUNK_SIZE)

    def _get_offset(self, i: int) -> int:
        return i * self.CHUNK_SIZE

    def _get_chunk_size(self, i: int) -> int:
        return self.LAST_CHUNK_SIZE if i == self.LAST_CHUNK_ID else self.CHUNK_SIZE

    def test_iter_id(self) -> None:
        for i, chunk_metadata in enumerate(self.tracker):
            self.assertEqual(i, chunk_metadata.id)

    def test_iter_offset(self) -> None:
        for i, chunk_metadata in enumerate(self.tracker):
            expected_offset = self._get_offset(i)
            self.assertEqual(expected_offset, chunk_metadata.offset)

    def test_iter_size(self) -> None:
        for i, chunk_metadata in enumerate(self.tracker):
            expected_size = self._get_chunk_size(i)
            self.assertEqual(expected_size, chunk_metadata.size)

    def test_iter_defaults_to_incomplete(self) -> None:
        for chunk_metadata in self.tracker:
            self.assertFalse(chunk_metadata.completed)

    def _set_complete(self, ids: list[int]) -> None:
        for i in ids:
            meta = ChunkMetadata(id=i, offset=self._get_offset(i), size=self._get_chunk_size(i), completed=False)
            self.tracker.set_complete(meta)

    def test_set_complete(self) -> None:
        ids = [1, 3]
        self._set_complete(ids)
        for meta in self.tracker:
            if meta.id in ids:
                self.assertTrue(meta.completed)
            else:
                self.assertFalse(meta.completed)

    @parameterized.expand(
        [
            ("first chunk", [0], 0.25),
            ("last chunk", [3], 0.25),
            ("odd chunks", [1, 3], 0.5),
            ("even chunks", [0, 2], 0.5),
            ("all chunks", [0, 1, 2, 3], 1.0),
        ]
    )
    def test_get_progress(self, _name: str, ids: list[int], expected_progress: float) -> None:
        self._set_complete(ids)
        actual_progress = self.tracker.get_progress()
        self.assertEqual(expected_progress, actual_progress)


class _TestIOError(ChunkedIOError):
    def __init__(self, message: str, raised_by: _TestIO):
        super().__init__(message)
        self.raised_by = raised_by

    async def recover(self) -> bool:
        try:
            await self.raised_by.renew()
            return True
        except:  # noqa: E722
            return False


class _TestIO(IDestination, ISource, object):
    """IDestination and ISource implementation for unit tests.
    expires_after: read or write operations will raise _TestIOError after the configured number of reads or
    writes. If expires_after is negative, _TestIO never expires and _TestIOError is never raised.
    """

    def __init__(self, content: bytes | None = None, expires_after: int = -1):
        if content is None:
            content = b""
        self._size = len(content)
        self._content = BytesIO(content)
        self._expires_after = expires_after
        self._n = 0
        self._io_lock = Lock()

    async def renew(self) -> None:
        self._n = 0

    @contextlib.contextmanager
    def _get_io_lock(self) -> None:
        with self._io_lock:
            if self._n == self._expires_after:
                raise _TestIOError(message="IO Expired", raised_by=self)
            self._n += 1
            yield

    async def write_chunk(self, offset: int, data: bytes) -> None:
        with self._get_io_lock():
            self._content.seek(offset)
            length = self._content.write(data)
            self._size = max(self._size, offset + length)

    async def read_chunk(self, offset: int, length: int) -> bytes:
        with self._get_io_lock():
            self._content.seek(offset)
            return self._content.read(length)

    async def get_size(self) -> int:
        return self._size

    def get_raw_content(self) -> bytes:
        self._content.seek(0)
        return self._content.read()


class TestChunkedIOManager(unittest.IsolatedAsyncioTestCase):
    FILE_SIZE = 15
    CHUNK_SIZE = 4
    DATA = b"\xaa\xaa\xaa\xaa\xbb\xbb\xbb\xbb\xcc\xcc\xcc\xcc\xdd\xdd\xdd"
    EXPIRES_AFTER = 2
    PART_DATA = DATA[: EXPIRES_AFTER * CHUNK_SIZE]

    def setUp(self) -> None:
        self.manager = ChunkedIOManager(
            retry=Retry(logger, backoff_method=BackoffLinear(0)), chunk_size=self.CHUNK_SIZE, max_workers=1
        )

    async def test_run_til_complete(self) -> None:
        # Create and verify source.
        source = _TestIO(content=self.DATA)
        self.assertEqual(self.DATA, source.get_raw_content())

        # Create and verify destination.
        destination = _TestIO()
        self.assertEqual(b"", destination.get_raw_content())

        # Transfer all data.
        await self.manager.run(source, destination)

        # Verify results.
        self.assertTrue(self.manager.is_complete())
        self.assertEqual(self.DATA, destination.get_raw_content())

    async def test_run_source_expires(self) -> None:
        # Create and verify source.
        source = _TestIO(content=self.DATA, expires_after=self.EXPIRES_AFTER)
        self.assertEqual(self.DATA, source.get_raw_content())

        # Create and verify destination.
        destination = _TestIO()
        self.assertEqual(b"", destination.get_raw_content())

        # Transfer data.
        with self.assertRaises(RetryError), mock.patch.object(source, "renew") as mock_renew:
            await self.manager.run(source, destination)
        self.assertEqual(2, mock_renew.call_count)

        # Verify results.
        self.assertFalse(self.manager.is_complete())
        self.assertEqual(self.PART_DATA, destination.get_raw_content())

    async def test_run_destination_expires(self) -> None:
        # Create and verify source.
        source = _TestIO(content=self.DATA)
        self.assertEqual(self.DATA, source.get_raw_content())

        # Create and verify destination.
        destination = _TestIO(expires_after=self.EXPIRES_AFTER)
        self.assertEqual(b"", destination.get_raw_content())

        # Transfer data.
        with self.assertRaises(RetryError), mock.patch.object(destination, "renew") as mock_renew:
            await self.manager.run(source, destination)
        self.assertEqual(2, mock_renew.call_count)

        # Verify results.
        self.assertFalse(self.manager.is_complete())
        self.assertEqual(self.PART_DATA, destination.get_raw_content())

    async def test_resume(self) -> None:
        # Create and verify source.
        source = _TestIO(content=self.DATA, expires_after=self.EXPIRES_AFTER)
        self.assertEqual(self.DATA, source.get_raw_content())

        # Create and verify destination.
        destination = _TestIO()
        self.assertEqual(b"", destination.get_raw_content())

        # Transfer data.
        with self.assertRaises(RetryError), mock.patch.object(source, "renew") as mock_renew:
            await self.manager.run(source, destination)
        self.assertEqual(2, mock_renew.call_count)

        # Verify results.
        self.assertFalse(self.manager.is_complete())
        self.assertEqual(self.PART_DATA, destination.get_raw_content())

        # Create new source and verify content.
        new_source = _TestIO(content=self.DATA)
        self.assertEqual(self.DATA, source.get_raw_content())

        # Finish data transfer.
        mock_read_chunk = mock.Mock(wraps=new_source.read_chunk)
        with mock.patch.object(new_source, "read_chunk", mock_read_chunk):
            await self.manager.run(new_source, destination)
        self.assertEqual(2, mock_read_chunk.call_count)

        # Verify results.
        self.assertTrue(self.manager.is_complete())
        self.assertEqual(self.DATA, destination.get_raw_content())

    async def test_resume_fails(self) -> None:
        # Create and verify source.
        source = _TestIO(content=self.DATA, expires_after=self.EXPIRES_AFTER)
        self.assertEqual(self.DATA, source.get_raw_content())

        # Create and verify destination.
        destination = _TestIO()
        self.assertEqual(b"", destination.get_raw_content())

        # Transfer data.
        with self.assertRaises(_TestIOError), mock.patch.object(source, "renew") as mock_renew:
            mock_renew.side_effect = ValueError("cannot renew")
            await self.manager.run(source, destination)
        self.assertEqual(1, mock_renew.call_count)

        # Verify results.
        self.assertFalse(self.manager.is_complete())



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys


class NoImport:
    """Simple context manager to prevent one or more named modules from being imported."""

    def __init__(self, *names: str) -> None:
        """
        :param names: The names of the modules to prevent from being imported.
        """
        self._names = names

    def __enter__(self) -> None:
        for name in self._names:
            # Set the module to None to prevent it from being imported.
            sys.modules[name] = None

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        for name in self._names:
            # Remove the module from sys.modules to clean up.
            del sys.modules[name]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Block Model API
=============


    The Block Model API provides the ability to manage and report on block models in your Evo workspaces. Enable your
    product with Evo connected workflows by integrating with the Seequent Block Model API.

    The Block Model API supports a range of sub-blocking options, and both full and partial updates to block models,
    including updates of specified columns and/or sub-volumes. Block models are versioned, and the service supports
    reporting on the material content of current or previous versions, and comparing the content between versions of
    a block model.

    For more information on the Block Model API, see [Overview](/docs/blockmodel/overview), or the API references here.


This code is generated from the OpenAPI specification for Block Model API.
API version: 1.16.1
"""

# Import endpoint apis.
from .api import ColumnOperationsApi, JobsApi, MetadataApi, OperationsApi, ReportsApi, UnitsApi, VersionsApi

__all__ = [
    "ColumnOperationsApi",
    "JobsApi",
    "MetadataApi",
    "OperationsApi",
    "ReportsApi",
    "UnitsApi",
    "VersionsApi",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from uuid import UUID

from evo.common import ICache, ResourceMetadata
from evo.common.io import Download
from evo.common.test_tools import TestWithDownloadHandler, UrlGenerator, utc_datetime

from .common import TEST_DATA


class MyResourceMetadata(ResourceMetadata):
    @property
    def url(self) -> str:
        return self.environment.hub_url + f"path/{self.name}"


class MyDownload(Download[MyResourceMetadata]):
    def __init__(self, metadata: MyResourceMetadata, url_generator: UrlGenerator) -> None:
        super().__init__()
        self._generator = url_generator
        self._metadata = metadata

    @property
    def metadata(self) -> MyResourceMetadata:
        return self._metadata

    @property
    def label(self) -> str:
        return "test_download"

    def _get_cache_location(self, cache: ICache) -> Path:
        return cache.get_location(self.metadata.environment, "downloads") / self.metadata.name

    async def get_download_url(self) -> str:
        return await self._generator.get_new_url()


class TestDownload(TestWithDownloadHandler):
    def setUp(self) -> None:
        super().setUp()
        self.setup_download_handler(TEST_DATA)
        self.metadata = MyResourceMetadata(
            environment=self.environment,
            id=UUID(int=4321),
            name="test_data_download.csv",
            created_at=utc_datetime(2024),
            created_by=None,
        )
        self.download = MyDownload(self.metadata, self.url_generator)

    async def test_to_path(self) -> None:
        """Test downloading a file to a path."""
        test_data_file = self.CACHE_DIR / "test_data_download.csv"
        self.assertFalse(test_data_file.exists())
        await self.download.download_to_path(test_data_file, self.transport)
        self.assert_download_requests(self.url_generator.current_url)
        self.assertTrue(test_data_file.exists())
        self.assertEqual(TEST_DATA, test_data_file.read_bytes())

    async def test_to_cache(self) -> None:
        """Test downloading a file to the cache."""
        expected_data_file = self.download._get_cache_location(self.cache)
        self.assertFalse(expected_data_file.exists())
        actual_data_file = await self.download.download_to_cache(self.cache, self.transport)
        self.assert_download_requests(self.url_generator.current_url)
        self.assertEqual(expected_data_file, actual_data_file)
        self.assertTrue(expected_data_file.exists())
        self.assertEqual(TEST_DATA, expected_data_file.read_bytes())



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import base64
import logging
import random
import unittest
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import datetime
from unittest import mock
from urllib.parse import urlencode

from parameterized import parameterized_class

from evo.common import RequestMethod
from evo.common.io.storage import BlockList, StorageBlock, StorageDestination
from evo.common.test_tools import TestWithUploadHandler, utc_datetime
from evo.common.utils import BackoffLinear, Retry

from .common import TEST_DATA, TestIDestination

logger = logging.getLogger(__name__)


@parameterized_class(
    [
        {"byte_offset": 0},
        {"byte_offset": 1},
        {"byte_offset": 1024},
        {"byte_offset": 4 * 1024 * 1024 * 1024 * 1024},
    ]
)
class TestStorageBlock(unittest.TestCase):
    byte_offset: int

    def setUp(self) -> None:
        self.block = StorageBlock(self.byte_offset)

    def test_byte_offset(self) -> None:
        """Test the block byte offset."""
        self.assertEqual(self.byte_offset, self.block.byte_offset)

    def test_id(self) -> None:
        """Test the block id generation."""
        expected_id = base64.b64encode(f"{self.byte_offset:032d}".encode("utf-8")).decode("utf-8")
        self.assertEqual(expected_id, self.block.id)

    def test_str(self) -> None:
        """Test the string representation of a block."""
        expected_str = f"<Latest>{self.block.id}</Latest>"
        self.assertEqual(expected_str, str(self.block))


@parameterized_class(
    [
        {"block_offsets": [0]},
        {"block_offsets": [1]},
        {"block_offsets": [i for i in range(0, 4 * 1024 * 1024, 512 * 1024)]},
    ]
)
class TestBlockList(unittest.IsolatedAsyncioTestCase):
    block_offsets: list[int]

    def setUp(self) -> None:
        self.block_list = BlockList()

    async def asyncSetUp(self) -> None:
        tasks = []
        for offset in random.sample(self.block_offsets, k=len(self.block_offsets)):
            tasks.append(self.block_list.add_block(offset))
        # Add a retry of the last task due to failure
        tasks.append(self.block_list.add_block(offset))
        self.sorted_block_ids = [StorageBlock(offset).id for offset in sorted(self.block_offsets)]
        self.randomized_block_ids = await asyncio.gather(*tasks)

    async def test_empty_list(self) -> None:
        """Test preparing an empty block list."""
        document = await BlockList().prepare()
        self.assertEqual('<?xml version="1.0" encoding="utf-8"?>\n<BlockList>\n</BlockList>', document.decode("utf-8"))

    async def test_populated_list(self) -> None:
        """Test preparing a populated block list."""
        if len(self.block_offsets) > 1:
            assert self.sorted_block_ids != self.randomized_block_ids
        document = await self.block_list.prepare()
        expected = "\n".join(
            [
                '<?xml version="1.0" encoding="utf-8"?>',
                "<BlockList>",
                *(f"  <Latest>{block_id}</Latest>" for block_id in self.sorted_block_ids),
                "</BlockList>",
            ]
        )
        self.assertEqual(expected, document.decode("utf-8"))


class TestStorageDestination(TestIDestination, TestWithUploadHandler):
    destination: StorageDestination

    def setup_destination(self) -> None:
        TestWithUploadHandler.setUp(self)
        self.destination = StorageDestination(self.url_generator.get_new_url, self.transport)

    def assert_put_block(self, offset: int, data: bytes) -> None:
        """Assert that a block was put to the server with the expected request format."""
        block = StorageBlock(offset)
        url = self.url_generator.current_url + "&" + urlencode({"comp": "block", "blockid": block.id})
        self.transport.assert_request_made(
            RequestMethod.PUT,
            url,
            headers={"Content-Length": str(len(data))},
            body=data,
        )

    def assert_put_block_list(self, block_list: bytes, with_url: str | None = None) -> None:
        """Assert that a block list was put to the server with the expected request format."""
        if with_url is None:
            with_url = self.url_generator.current_url
        self.transport.assert_any_request_made(
            RequestMethod.PUT,
            with_url + "&" + urlencode({"comp": "blocklist"}),
            headers={
                "Content-Type": "text/plain; charset=UTF-8",
                "Content-Length": str(len(block_list)),
            },
            body=block_list,
        )

    @asynccontextmanager
    async def ctx_test_write_chunk(self, offset: int, data: bytes) -> AsyncIterator[None]:
        async with self.destination:
            self.transport.assert_no_requests()
            self.assertEqual(1, self.url_generator.n_calls)
            yield
            self.assert_put_block(offset, data)
            self.assertEqual(1, self.url_generator.n_calls)

    async def test_commit(self) -> AsyncIterator[None]:
        """Test committing a file to storage submits the expected block list."""
        async with self.destination:
            chunks_with_offsets = await self.write_whole_file()
            self.transport.reset_mock()
            self.assertEqual(1, self.url_generator.n_calls)

            expected_block_list = BlockList()
            tasks = [expected_block_list.add_block(offset) for offset, _ in chunks_with_offsets]
            await asyncio.gather(*tasks)

            await self.destination.commit()
            self.assert_put_block_list(await expected_block_list.prepare())
            self.assertEqual(1, self.url_generator.n_calls)

        expected_data = b"".join(data for _, data in sorted(chunks_with_offsets))
        actual_data = await self.handler.get_committed()
        self.assertEqual(expected_data, actual_data)

    @mock.patch("evo.common.io.http.datetime", spec=datetime)
    async def test_commit_retries(self, mock_datetime: mock.Mock) -> None:
        """Test that committing a file to storage retries on auth failure."""
        mock_datetime.now.return_value = utc_datetime(2024, 7, 12, hour=12, minute=0)
        async with self.destination:
            chunks_with_offsets = await self.write_whole_file()
            self.transport.reset_mock()
            self.assertEqual(1, self.url_generator.n_calls)

            expected_block_list = BlockList()
            tasks = [expected_block_list.add_block(offset) for offset, _ in chunks_with_offsets]
            await asyncio.gather(*tasks)

            mock_datetime.now.return_value = utc_datetime(2024, 7, 12, hour=12, minute=5)
            expired_url = self.url_generator.current_url
            with self.handler.expired_url(expired_url):
                await self.destination.commit(
                    # Use a retry strategy that doesn't wait between attempts.
                    retry=Retry(logger=logger, max_attempts=3, backoff_method=BackoffLinear(0.0)),
                )

            expected_document = await expected_block_list.prepare()
            self.assert_put_block_list(expected_document, with_url=expired_url)
            self.assert_put_block_list(expected_document, with_url=self.url_generator.current_url)
            self.assertEqual(2, self.url_generator.n_calls)

    async def test_upload_file(self) -> None:
        test_data_file = self.CACHE_DIR / "test_data_upload.csv"
        test_data_file.write_bytes(TEST_DATA)
        await StorageDestination.upload_file(test_data_file, self.url_generator.get_new_url, self.transport)
        uploaded_data = await self.handler.get_committed()
        self.assertEqual(test_data_file.read_bytes(), uploaded_data)


# Delete base test classes to prevent discovery by unittest.
del TestIDestination



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import contextlib
import hashlib
import hmac
import json
import unittest
from base64 import urlsafe_b64encode
from collections.abc import Iterator
from datetime import datetime
from typing import Any, Literal
from unittest import mock
from urllib.parse import urlencode

from aiohttp.client import ClientSession

from evo.common import RequestMethod
from evo.common.test_tools import TestTransport, utc_datetime
from evo.oauth import AnyScopes, EvoScopes, OAuthConnector
from evo.oauth.data import AccessToken

ACCESS_TOKEN = "TestAccessToken"
BASE_URI = "https://auth.unittest.test"
CLIENT_ID = "TestClientID"
CLIENT_SECRET = "TestClientSecret"
REFRESH_TOKEN = "TestRefreshToken"
EXPIRES_IN = 3600
ISSUED_AT = utc_datetime(2024, 4, 1)

REDIRECT_URL = "http://localhost:1234"
STATE_TOKEN = "TestStateToken"
VERIFIER_TOKEN = "TestVerifierToken"
AUTHORIZATION_CODE = "TestAuthorizationCode"
OAUTH_SCOPES = EvoScopes.all_evo


def _jwt_encode(**data: Any) -> str:
    return urlsafe_b64encode(json.dumps(data).encode("utf-8")).decode("utf-8").rstrip("=")


def _get_jwt(issuer: str, subject: str, audience: str | list[str], expiry: int, issued_at: int) -> str:
    header = _jwt_encode(typ="JWT", alg="HS256")
    payload = _jwt_encode(iss=issuer, sub=subject, aud=audience, exp=expiry, iat=issued_at)
    token = f"{header}.{payload}"
    digest = hmac.new(b"secret", token.encode("utf-8"), "sha256").digest()
    return token + "." + urlsafe_b64encode(digest).decode("utf-8").rstrip("=")


def get_user_access_token(
    access_token: str = ACCESS_TOKEN,
    refresh_token: str | None = REFRESH_TOKEN,
    expires_in: int = EXPIRES_IN,
    issued_at: datetime = ISSUED_AT,
) -> AccessToken:
    return AccessToken(
        token_type="Bearer",
        access_token=access_token,
        refresh_token=refresh_token,
        expires_in=expires_in,
        issued_at=issued_at,
    )


def get_access_token(
    access_token: str = ACCESS_TOKEN,
    refresh_token: str | None = REFRESH_TOKEN,
    expires_in: int = EXPIRES_IN,
    issued_at: datetime = ISSUED_AT,
) -> AccessToken:
    return AccessToken(
        token_type="Bearer",
        access_token=access_token,
        refresh_token=refresh_token,
        expires_in=expires_in,
        issued_at=issued_at,
    )


@contextlib.contextmanager
def patch_urlsafe_tokens(state: str = STATE_TOKEN, verifier: str = VERIFIER_TOKEN) -> Iterator[None]:
    with mock.patch("secrets.token_urlsafe", side_effect=[state, verifier]):
        yield


async def get_redirect(state: str = STATE_TOKEN, code: str = AUTHORIZATION_CODE) -> str:
    async with ClientSession() as session:
        async with session.get(REDIRECT_URL, params={"state": state, "code": code}) as response:
            response.raise_for_status()
            return await response.text()


@contextlib.contextmanager
def patch_webbrowser_open(authenticate: bool) -> Iterator[mock.Mock]:
    """Patch the `webbrowser.open` function and optionally authenticate the user."""

    def webbrowser_open(_url: str) -> None:
        if authenticate:
            asyncio.ensure_future(get_redirect(state=STATE_TOKEN, code=AUTHORIZATION_CODE))

    with mock.patch("webbrowser.open", side_effect=webbrowser_open) as mock_open:
        yield mock_open


class TestWithOAuthConnector(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.transport = TestTransport(base_url=BASE_URI)
        self.connector = OAuthConnector(
            self.transport,
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET,
            base_uri=BASE_URI,
        )

    def get_expected_auth_url(
        self, state: str = STATE_TOKEN, verifier: str = VERIFIER_TOKEN, scopes: AnyScopes = EvoScopes.default
    ) -> str:
        base_url = self.connector.base_uri + self.connector.endpoint("authorize")

        # urlsafe base64 encoded sha256 hash of the verifier token.
        expect_challenge = (
            urlsafe_b64encode(
                hashlib.sha256(verifier.encode()).digest(),
            )
            .decode()
            .strip("=")
        )

        query = {
            "response_type": "code",
            "client_id": self.connector.client_id,
            "redirect_uri": REDIRECT_URL,
            "state": state,
            "scope": str(scopes),
            "code_challenge": expect_challenge,
            "code_challenge_method": "S256",
        }
        return base_url + "?" + urlencode(sorted(query.items()))

    def assert_fetched_token(self, path: str = "/connect/token", **data: str) -> None:
        """Assert that a token was fetched with the given data"""
        self.transport.assert_any_request_made(
            RequestMethod.POST,
            path,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            post_params=list(data.items()),
        )


class MockOAuthConnector(mock.AsyncMock):
    def __init__(self, base_uri=BASE_URI, client_id=CLIENT_ID, client_secret=CLIENT_SECRET, **kwargs):
        super().__init__(kwargs=kwargs)
        self._base_uri = base_uri
        self._client_id = client_id
        self._client_secret = client_secret
        self._config = None

    @property
    def base_uri(self):
        return BASE_URI

    @property
    def client_id(self):
        return self._client_id

    def endpoint(self, endpoint_type: Literal["authorize", "token"]) -> str:
        """
        Returns the relevant OAuth endpoint by endpoint type. Possible values: "authorize", "token".
        """
        match endpoint_type:
            case "authorize":
                return "/connect/authorize"
            case "token":
                return "/connect/token"


class TestWithMockOAuthConnector(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.transport = TestTransport(base_url=BASE_URI)
        self.connector = MockOAuthConnector(spec=OAuthConnector)

    def get_expected_auth_url(
        self, state: str = STATE_TOKEN, verifier: str = VERIFIER_TOKEN, scopes: AnyScopes = EvoScopes.default
    ) -> str:
        base_url = self.connector.base_uri + "/connect/authorize"

        # urlsafe base64 encoded sha256 hash of the verifier token.
        expect_challenge = (
            urlsafe_b64encode(
                hashlib.sha256(verifier.encode()).digest(),
            )
            .decode()
            .strip("=")
        )

        query = {
            "response_type": "code",
            "client_id": self.connector.client_id,
            "redirect_uri": REDIRECT_URL,
            "state": state,
            "scope": str(scopes),
            "code_challenge": expect_challenge,
            "code_challenge_method": "S256",
        }
        return base_url + "?" + urlencode(sorted(query.items()))

    def assert_fetched_token(self, path: str = "/connect/token", **data: str) -> None:
        """Assert that a token was fetched with the given data"""
        self.transport.assert_any_request_made(
            RequestMethod.POST,
            path,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            post_params=list(data.items()),
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from uuid import UUID

from parameterized import param, parameterized

from evo.common.data import RequestMethod
from evo.common.test_tools import MockResponse, TestWithConnector
from evo.discovery import DiscoveryAPIClient, Hub, Organization

from ..data import load_test_data


def _sample_data_as_response_content(sample_data: dict) -> str:
    return json.dumps({"discovery": sample_data})


def _get_service_access_services(org: dict, hub: dict, service_access: list[dict]) -> list[str]:
    return next(
        service_access
        for service_access in service_access
        if service_access["hub_code"] == hub["code"] and service_access["org_id"] == org["id"]
    )["services"]


def _sample_data_as_expected_orgs(sample_data: dict) -> list[Organization]:
    all_hubs = sorted(sample_data["hubs"], key=lambda h: h["display_name"])
    org_data = sorted(sample_data["organizations"], key=lambda o: o["display_name"])
    service_access = {(sa["org_id"], sa["hub_code"]) for sa in sample_data["service_access"]}
    return [
        Organization(
            id=UUID(org["id"]),
            display_name=org["display_name"],
            hubs=tuple(
                Hub(
                    url=hub["url"],
                    code=hub["code"],
                    display_name=hub["display_name"],
                    services=tuple(_get_service_access_services(org, hub, sample_data["service_access"])),
                )
                for hub in all_hubs
                if (org["id"], hub["code"]) in service_access
            ),
        )
        for org in org_data
    ]


class TestDiscoveryAPIClient(TestWithConnector):
    def setUp(self) -> None:
        super().setUp()
        self.discovery_client = DiscoveryAPIClient(self.connector)
        self.transport.request.return_value = MockResponse(status_code=500)

    async def test_list_organizations_default_service_code(self) -> None:
        """Test a successful get organizations request with the default service code."""
        test_data = load_test_data("successful_service_discovery.json")
        test_data_text = json.dumps(test_data)
        with self.transport.set_http_response(
            status_code=200, content=test_data_text, headers={"Content-Type": "application/json"}
        ):
            await self.discovery_client.list_organizations()
        expected_path = "/evo/identity/v2/discovery?service=evo"
        self.assert_request_made(method=RequestMethod.GET, path=expected_path)

    async def test_list_organizations_with_custom_service_codes(self) -> None:
        """Test a successful get organizations request with custom service codes."""
        test_data = load_test_data("successful_service_discovery.json")
        test_data_text = json.dumps(test_data)
        with self.transport.set_http_response(
            status_code=200, content=test_data_text, headers={"Content-Type": "application/json"}
        ):
            await self.discovery_client.list_organizations(["service0", "service1", "service2"])
        expected_path = "/evo/identity/v2/discovery?service=service0&service=service1&service=service2"
        self.assert_request_made(method=RequestMethod.GET, path=expected_path)

    @parameterized.expand([param(**scenario) for scenario in load_test_data("service_discovery_data.json")])
    async def test_list_organizations(self, scenario: str, sample_data: dict) -> None:
        expected_orgs = _sample_data_as_expected_orgs(sample_data)
        with self.transport.set_http_response(
            status_code=200,
            content=_sample_data_as_response_content(sample_data),
            headers={"Content-Type": "application/json"},
        ):
            actual_orgs = await self.discovery_client.list_organizations()
        self.assertListEqual(expected_orgs, actual_orgs)

        # Test that the organizations are sorted alphanumerically.
        self.assertListEqual(sorted(actual_orgs, key=lambda o: o.display_name), actual_orgs)

        # Test that the hubs are sorted alphanumerically.
        for org in actual_orgs:
            self.assertListEqual(sorted(org.hubs, key=lambda h: h.display_name), list(org.hubs))

    def test_organization_is_hashable(self) -> None:
        """Test that the Organization dataclass is hashable."""
        org = Organization(
            id=UUID("12345678-1234-5678-1234-567812345678"),
            display_name="Hashable org name",
            hubs=(
                Hub(
                    url="https://example.com",
                    code="test_hub",
                    display_name="Test Hub",
                    services=("service1", "service2"),
                ),
            ),
        )
        try:
            hash_value = hash(org)
            self.assertIsInstance(hash_value, int)
        except TypeError:
            self.fail("Organization dataclass is not hashable")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import uuid
from pathlib import Path

from evo.common import APIConnector, Environment, ICache
from evo.common.io import Download, Upload
from evo.common.io.exceptions import RenewalError

from ._utils import extract_payload
from .endpoints.api import JobsApi
from .endpoints.models import JobPendingPayload, JobStatus, QueryDownload, QueryResult
from .exceptions import UnknownJobPayload

_CACHE_SCOPE = "blockmodel"


class BlockModelDownload(Download[QueryResult]):
    def __init__(
        self,
        connector: APIConnector,
        environment: Environment,
        query_result: QueryResult,
        job_id: uuid.UUID,
        initial_url: str | None = None,
    ) -> None:
        self._jobs_api = JobsApi(connector)
        self._environment = environment
        self._query_result = query_result
        self._job_id = job_id
        self._mutex = asyncio.Lock()
        self._initial_url = initial_url

    @property
    def label(self) -> str:
        """The label of the resource to be downloaded."""
        return "Block Model"

    @property
    def metadata(self) -> QueryResult:
        """The metadata of the resource to be downloaded."""
        return self._query_result

    def _get_cache_location(self, cache: ICache) -> Path:
        """Generate the cache location for the resource to be downloaded.

        :param cache: The cache to resolve the cache location.

        :returns: The cache location.
        """
        return cache.get_location(environment=self._environment, scope=_CACHE_SCOPE) / str(self._job_id)

    async def get_download_url(self) -> str:
        """Generate a URL that will be used to download the resource.

        This method may be called multiple times to generate a new URL if the last URL expires.

        :returns: The download URL.
        """
        async with self._mutex:
            if (url := self._initial_url) is not None:
                # Clear the initial URL so it can't be reused.
                self._initial_url = None
                return url

        response = await self._jobs_api.get_job_status(
            job_id=str(self._job_id),
            workspace_id=str(self._environment.workspace_id),
            org_id=str(self._environment.org_id),
            bm_id=str(self._query_result.bm_uuid),
        )
        if response.job_status != JobStatus.COMPLETE:
            raise RenewalError("Job is not complete")
        try:
            payload = extract_payload(self._job_id, response, QueryDownload)
        except UnknownJobPayload as e:
            raise RenewalError(str(e)) from e
        return str(payload.download_url)


class BlockModelUpload(Upload):
    def __init__(
        self,
        connector: APIConnector,
        environment: Environment,
        bm_uuid: uuid.UUID,
        job_id: uuid.UUID,
        initial_url: str | None = None,
    ) -> None:
        self._jobs_api = JobsApi(connector)
        self._environment = environment
        self._bm_uuid = bm_uuid
        self._job_id = job_id
        self._mutex = asyncio.Lock()
        self._initial_url = initial_url

    @property
    def label(self) -> str:
        """The label of the resource to be uploaded."""
        return "Block Model"

    async def get_upload_url(self) -> str:
        """Generate a URL that will be used to upload the resource.

        This method may be called multiple times to generate a new URL if the last URL expires.

        :returns: The upload URL.

        :raises DataExistsError: if the resource already exists in the target service.
        """
        async with self._mutex:
            if (url := self._initial_url) is not None:
                # Clear the initial URL so it can't be reused.
                self._initial_url = None
                return url

        response = await self._jobs_api.get_job_status(
            job_id=str(self._job_id),
            workspace_id=str(self._environment.workspace_id),
            org_id=str(self._environment.org_id),
            bm_id=str(self._bm_uuid),
        )
        if response.job_status != JobStatus.PENDING_UPLOAD:
            raise RenewalError("Job is not in the pending upload state")
        try:
            payload = extract_payload(self._job_id, response, JobPendingPayload)
        except UnknownJobPayload as e:
            raise RenewalError(str(e)) from e
        return str(payload.upload_url)


def get_cache_location_for_upload(cache: ICache, environment: Environment, job_id: uuid.UUID) -> Path:
    return cache.get_location(environment=environment, scope=_CACHE_SCOPE) / str(job_id)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from datetime import datetime
from typing import Iterable
from unittest import mock

import pyarrow

from evo.blockmodels import BlockModelAPIClient
from evo.blockmodels.data import RegularGridDefinition
from evo.blockmodels.endpoints import models
from evo.blockmodels.endpoints.models import JobResponse, JobStatus, RotationAxis
from evo.blockmodels.exceptions import CacheNotConfiguredException, JobFailedException
from evo.common import ServiceUser
from evo.common.data import HTTPHeaderDict, RequestMethod
from evo.common.test_tools import BASE_URL, MockResponse, TestWithConnector, TestWithStorage
from utils import JobPollingRequestHandler

BM_UUID = uuid.uuid4()
GOOSE_UUID = uuid.uuid4()
GOOSE_VERSION_ID = "2"
DATE = datetime(2021, 1, 1)
MODEL_USER = models.UserInfo(email="test@test.com", name="Test User", id=uuid.uuid4())
USER = ServiceUser.from_model(MODEL_USER)
BM_BBOX = models.BBoxXYZ(
    x_minmax=models.FloatRange(min=0, max=10),
    y_minmax=models.FloatRange(min=0, max=10),
    z_minmax=models.FloatRange(min=0, max=10),
)
GRID_DEFINITION = RegularGridDefinition(
    model_origin=[0, 0, 0], rotations=[(RotationAxis.x, 20)], n_blocks=[10, 10, 10], block_size=[1, 1, 1]
)


def _mock_create_result(environment) -> models.BlockModelAndJobURL:
    return models.BlockModelAndJobURL(
        bbox=BM_BBOX,
        block_rotation=[models.Rotation(axis=RotationAxis.x, angle=20)],
        bm_uuid=BM_UUID,
        name="Test BM",
        description="Test Block Model",
        coordinate_reference_system="EPSG:4326",
        size_unit_id="m",
        workspace_id=environment.workspace_id,
        org_uuid=environment.org_id,
        model_origin=models.Location(x=0, y=0, z=0),
        normalized_rotation=[0, 20, 0],
        size_options=models.SizeOptionsRegular(
            n_blocks=models.Size3D(nx=10, ny=10, nz=10),
            block_size=models.BlockSize(x=1, y=1, z=1),
        ),
        geoscience_object_id=GOOSE_UUID,
        created_at=DATE,
        created_by=MODEL_USER,
        last_updated_at=DATE,
        last_updated_by=MODEL_USER,
        job_url=f"{BASE_URL}/jobs/{uuid.uuid4()}",
    )


def _mock_version(
    version_id: int, version_uuid: uuid.UUID, goose_version_id: str, bbox=None, columns: Iterable[models.Column] = ()
) -> models.Version:
    return models.Version(
        base_version_id=None if version_id == 1 else version_id - 1,
        bbox=bbox,
        bm_uuid=BM_UUID,
        comment="",
        created_at=DATE,
        created_by=MODEL_USER,
        geoscience_version_id=goose_version_id,
        mapping=models.Mapping(columns=list(columns)),
        parent_version_id=version_id - 1,
        version_id=version_id,
        version_uuid=version_uuid,
    )


FIRST_VERSION = _mock_version(1, uuid.uuid4(), "2")

UPDATE_RESULT = models.UpdateWithUrl(
    changes=models.UpdateDataLiteOutput(
        # We don't look at these values, so we can just set them to empty
        columns=models.UpdateColumnsLiteOutput(new=[], update=[], rename=[], delete=[])
    ),
    version_uuid=FIRST_VERSION.version_uuid,
    job_uuid=uuid.uuid4(),
    job_url=f"{BASE_URL}/jobs/{uuid.uuid4()}",
    upload_url=f"{BASE_URL}/upload/{uuid.uuid4()}",
)

INITIAL_DATA = pyarrow.table(
    {"i": [1, 2, 3], "j": [4, 5, 6], "k": [7, 8, 9], "col1": ["A", "B", "B"], "col2": [4.5, 5.3, 6.2]}
)


class CreateRequestHandler(JobPollingRequestHandler):
    def __init__(
        self,
        create_result: models.BlockModelAndJobURL,
        job_response: JobResponse,
        update_result: models.UpdateWithUrl | None = None,
        update_job_response: JobResponse | None = None,
        pending_request: int = 0,
    ) -> None:
        super().__init__(job_response, pending_request)
        self._create_result = create_result
        self._update_result = update_result
        self._update_job_response = update_job_response

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        match method:
            case RequestMethod.POST:
                return MockResponse(status_code=201, content=self._create_result.model_dump_json())
            case RequestMethod.PATCH:
                if self._update_result is None:
                    return self.not_found()
                self._job_response = self._update_job_response
                return MockResponse(status_code=202, content=self._update_result.model_dump_json())
            case RequestMethod.GET:
                return self.job_poll()
            case _:
                return self.not_found()


class TestCreateBlockModel(TestWithConnector, TestWithStorage):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithStorage.setUp(self)
        self.bms_client = BlockModelAPIClient(connector=self.connector, environment=self.environment, cache=self.cache)
        self.bms_client_without_cache = BlockModelAPIClient(connector=self.connector, environment=self.environment)

    @property
    def base_path(self) -> str:
        return f"blockmodel/orgs/{self.environment.org_id}/workspaces/{self.environment.workspace_id}"

    async def test_create_block_model(self) -> None:
        self.transport.set_request_handler(
            CreateRequestHandler(
                create_result=_mock_create_result(self.environment),
                job_response=JobResponse(
                    job_status=JobStatus.COMPLETE,
                    payload=FIRST_VERSION,
                ),
            )
        )
        bm, version = await self.bms_client.create_block_model(
            name="Test BM",
            description="Test Block Model",
            grid_definition=GRID_DEFINITION,
            object_path="test/path",
            coordinate_reference_system="EPSG:4326",
            size_unit_id="m",
        )
        self.assertEqual(bm.id, BM_UUID)
        self.assertEqual(bm.geoscience_object_id, GOOSE_UUID)
        self.assertEqual(bm.name, "Test BM")
        self.assertEqual(bm.description, "Test Block Model")
        self.assertEqual(bm.grid_definition.model_origin, [0, 0, 0])
        self.assertEqual(bm.grid_definition.rotations, [(RotationAxis.x, 20)])
        self.assertIsInstance(bm.grid_definition, RegularGridDefinition)
        self.assertEqual(bm.grid_definition.n_blocks, [10, 10, 10])
        self.assertEqual(bm.grid_definition.block_size, [1, 1, 1])
        self.assertEqual(bm.coordinate_reference_system, "EPSG:4326")
        self.assertEqual(bm.size_unit_id, "m")
        self.assertEqual(bm.bbox, BM_BBOX)
        self.assertEqual(bm.created_at, DATE)
        self.assertEqual(bm.created_by, USER)
        self.assertEqual(bm.last_updated_at, DATE)
        self.assertEqual(bm.last_updated_by, USER)

        self.assertEqual(version.bm_uuid, BM_UUID)
        self.assertEqual(version.version_id, 1)
        self.assertEqual(version.version_uuid, FIRST_VERSION.version_uuid)
        self.assertEqual(version.parent_version_id, 0)
        self.assertEqual(version.base_version_id, None)
        self.assertEqual(version.geoscience_version_id, "2")
        self.assertEqual(version.bbox, None)
        self.assertEqual(version.comment, "")
        self.assertEqual(version.created_at, DATE)
        self.assertEqual(version.created_by, USER)
        self.assertEqual(version.columns, [])

    async def test_create_block_model_with_data(self) -> None:
        second_version = _mock_version(
            2,
            uuid.uuid4(),
            "3",
            models.BBox(
                i_minmax=models.IntRange(min=1, max=3),
                j_minmax=models.IntRange(min=4, max=6),
                k_minmax=models.IntRange(min=7, max=9),
            ),
            columns=[
                models.Column(col_id=str(uuid.uuid4()), title="col1", data_type=models.DataType.Utf8),
                models.Column(col_id=str(uuid.uuid4()), title="col2", data_type=models.DataType.Float64),
            ],
        )

        self.transport.set_request_handler(
            CreateRequestHandler(
                create_result=_mock_create_result(self.environment),
                job_response=JobResponse(
                    job_status=JobStatus.COMPLETE,
                    payload=FIRST_VERSION,
                ),
                update_result=UPDATE_RESULT,
                update_job_response=JobResponse(
                    job_status=JobStatus.COMPLETE,
                    payload=second_version,
                ),
            )
        )
        with (
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_destination.upload_file = mock.AsyncMock()
            bm, version = await self.bms_client.create_block_model(
                name="Test BM",
                description="Test Block Model",
                grid_definition=GRID_DEFINITION,
                object_path="test/path",
                coordinate_reference_system="EPSG:4326",
                size_unit_id="m",
                initial_data=INITIAL_DATA,
                units={"col2": "g/t"},
            )
            mock_destination.upload_file.assert_called_once()

            # Assert that the correct columns are part of the update
            expected_update_body = models.UpdateDataLiteInput(
                columns=models.UpdateColumnsLiteInput(
                    new=[
                        models.ColumnLite(
                            title="col1",
                            data_type=models.DataType.Utf8,
                            unit_id=None,
                        ),
                        models.ColumnLite(
                            title="col2",
                            data_type=models.DataType.Float64,
                            unit_id="g/t",
                        ),
                    ],
                    update=[],
                    rename=[],
                    delete=[],
                ),
                update_type=models.UpdateType.replace,
            )
            self.transport.assert_any_request_made(
                method=RequestMethod.PATCH,
                path=f"{self.base_path}/block-models/{BM_UUID}/blocks",
                body=expected_update_body.model_dump(mode="json", exclude_unset=True),
                headers={
                    "Authorization": "Bearer <not-a-real-token>",
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )
        self.assertEqual(bm.id, BM_UUID)

        self.assertEqual(version.bm_uuid, BM_UUID)
        self.assertEqual(version.version_id, 2)
        self.assertEqual(version.version_uuid, second_version.version_uuid)
        self.assertEqual(version.parent_version_id, 1)
        self.assertEqual(version.base_version_id, 1)
        self.assertEqual(version.geoscience_version_id, "3")
        self.assertEqual(version.bbox, second_version.bbox)
        self.assertEqual(version.comment, "")
        self.assertEqual(version.created_at, DATE)
        self.assertEqual(version.created_by, USER)
        self.assertEqual(version.columns, second_version.mapping.columns)

    async def test_create_block_model_no_cache(self) -> None:
        with self.assertRaises(CacheNotConfiguredException):
            await self.bms_client_without_cache.create_block_model(
                name="Test BM",
                description="Test Block Model",
                grid_definition=GRID_DEFINITION,
                object_path="test/path",
                coordinate_reference_system="EPSG:4326",
                size_unit_id="m",
                initial_data=INITIAL_DATA,
            )

    async def test_create_block_model_job_failed(self) -> None:
        self.transport.set_request_handler(
            CreateRequestHandler(
                create_result=_mock_create_result(self.environment),
                job_response=JobResponse(
                    job_status=JobStatus.COMPLETE,
                    payload=FIRST_VERSION,
                ),
                update_result=UPDATE_RESULT,
                update_job_response=JobResponse(
                    job_status=JobStatus.FAILED,
                    payload=models.JobErrorPayload(
                        detail="Update Job failed",
                        status=500,
                        title="Update Job failed",
                        type="https://seequent.com/error-codes/block-model-service/job/internal-error",
                    ),
                ),
            )
        )
        with (
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_destination.upload_file = mock.AsyncMock()
            with self.assertRaises(JobFailedException):
                await self.bms_client.create_block_model(
                    name="Test BM",
                    description="Test Block Model",
                    grid_definition=GRID_DEFINITION,
                    object_path="test/path",
                    coordinate_reference_system="EPSG:4326",
                    size_unit_id="m",
                    initial_data=INITIAL_DATA,
                )
            mock_destination.upload_file.assert_called_once()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# generated by datamodel-codegen:
#   filename:  spec.yaml

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Annotated
from uuid import UUID

from pydantic import Field, RootModel, StrictBool, StrictFloat, StrictInt, StrictStr

from .._model_config import CustomBaseModel


class CoordinateItem(RootModel[list[StrictFloat]]):
    root: Annotated[list[StrictFloat], Field(max_length=2, min_length=2, title="GeoJSON Point")]
    """
    A WGS84 coordinate pair [longitude, latitude]. The longitude must come before the latitude, in order to be considered a valid GeoJSON Point
    """


class Coordinate(RootModel[list[CoordinateItem]]):
    root: Annotated[
        list[CoordinateItem],
        Field(max_length=5, min_length=5, title="GeoJSON LinearRing"),
    ]
    """
    A list of GeoJSON points defining a closed linear ring. In order for the ring to be closed, the last point must be the same as the first point. Only a rectangle is supported, so the ring must be defined by 5 points.
    """


class DataDownloadUrl(CustomBaseModel):
    download_url: Annotated[StrictStr | None, Field(title="Download Url")] = None
    id: Annotated[UUID, Field(title="Id")]
    name: Annotated[StrictStr, Field(title="Name")]


class DataUploadRequestBody(CustomBaseModel):
    name: Annotated[StrictStr, Field(title="Name")]
    """
    SHA256 or UUID "name" for the data blob. SHA256 is preferred as it allows other clients to deduplicate data.
    """


class DataUploadResponseBody(CustomBaseModel):
    exists: Annotated[StrictBool, Field(title="Exists")]
    id: Annotated[UUID, Field(title="Id")]
    """
    Server-generated unique identifier for the data
    """
    name: Annotated[StrictStr, Field(title="Name")]
    """
    client-provided name for the data (SHA256 or UUID)
    """
    upload_url: Annotated[StrictStr | None, Field(title="Upload Url")] = None


class GeoJSONType(Enum):
    Polygon = "Polygon"


class GeoscienceObject(CustomBaseModel):
    """
    The geoscience object with the additional required schema properties.

    This object is largely defined by the schemas in the geoscience-object-schemas repository.
    There are some default, required values listed here explicitly.
    """

    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the geoscience object.
    """
    schema_: Annotated[StrictStr, Field(alias="schema", title="Schema")]
    """
    The geoscience object schema.
    """
    uuid: Annotated[UUID | None, Field(title="Uuid")] = None
    """
    Unique object UUID.
    """


class LatestObjectVersionIdResponse(CustomBaseModel):
    object_id: Annotated[StrictStr, Field(title="Object Id")]
    version_id: Annotated[StrictStr | None, Field(title="Version Id")] = None


class ListObjectsResponseLinks(CustomBaseModel):
    next: Annotated[StrictStr | None, Field(title="Next")] = None
    prev: Annotated[StrictStr | None, Field(title="Prev")] = None


class ListedObjectLinks(CustomBaseModel):
    download: Annotated[StrictStr, Field(title="Download")]


class MetadataUpdateBody(CustomBaseModel):
    stage_id: Annotated[UUID | None, Field(title="Stage Id")] = None


class ObjectResponseLinks(CustomBaseModel):
    data: Annotated[list[DataDownloadUrl], Field(title="Data")]
    download: Annotated[StrictStr, Field(title="Download")]


class ObjectVersionResponseLinks(CustomBaseModel):
    download: Annotated[StrictStr, Field(title="Download")]


class RoleEnum(Enum):
    owner = "owner"
    editor = "editor"
    viewer = "viewer"


class StageListItem(CustomBaseModel):
    name: Annotated[StrictStr, Field(title="Name")]
    stage_id: Annotated[UUID, Field(title="Stage Id")]


class StageResponse(CustomBaseModel):
    name: Annotated[StrictStr, Field(title="Name")]
    stage_id: Annotated[UUID, Field(title="Stage Id")]


class UpdateGeoscienceObject(CustomBaseModel):
    """
    The geoscience object with the additional required schema properties.

    The uuid field must be set when updating a geoscience object.
    This object is largely defined by the schemas in the geoscience-object-schemas repository.
    There are some default, required values listed here explicitly.
    """

    name: Annotated[StrictStr, Field(title="Name")]
    """
    The name of the geoscience object.
    """
    schema_: Annotated[StrictStr, Field(alias="schema", title="Schema")]
    """
    The geoscience object schema.
    """
    uuid: Annotated[UUID, Field(title="Uuid")]
    """
    Unique object UUID.
    """


class User(CustomBaseModel):
    """
    User details.
    """

    email: Annotated[StrictStr | None, Field(title="Email")] = None
    id: Annotated[UUID, Field(title="Id")]
    name: Annotated[StrictStr | None, Field(title="Name")] = None


class BoundingBox(CustomBaseModel):
    """
    The bounding box for a geoscience object, represented by a GeoJSON polygon.
    """

    coordinates: Annotated[
        list[Coordinate],
        Field(max_length=1, min_length=1, title="GeoJSON Polygon Coordinates"),
    ]
    """
    A list containing one exterior ring of a GeoJSON polygon
    """
    type: GeoJSONType = "Polygon"


class GeoscienceObjectVersion(CustomBaseModel):
    """
    Geoscience object version.
    """

    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: User | None = None
    etag: Annotated[StrictStr, Field(title="Etag")]
    links: ObjectVersionResponseLinks
    stage: StageResponse | None = None
    version_id: Annotated[StrictStr, Field(title="Version Id")]


class GetObjectResponse(CustomBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: User | None = None
    deleted_at: Annotated[datetime | None, Field(title="Deleted At")] = None
    deleted_by: User | None = None
    etag: Annotated[StrictStr, Field(title="Etag")]
    geojson_bounding_box: BoundingBox | None = None
    links: ObjectResponseLinks
    modified_at: Annotated[datetime, Field(title="Modified At")]
    modified_by: User | None = None
    object: GeoscienceObject
    object_id: Annotated[UUID, Field(title="Object Id")]
    object_path: Annotated[StrictStr | None, Field(title="Object Path")] = None
    stage: StageResponse | None = None
    version_id: Annotated[StrictStr, Field(title="Version Id")]
    versions: Annotated[list[GeoscienceObjectVersion] | None, Field(title="Versions")] = None


class ListStagesResponse(CustomBaseModel):
    stages: Annotated[list[StageListItem], Field(title="Stages")]


class ListedObject(CustomBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: User | None = None
    deleted_at: Annotated[datetime | None, Field(title="Deleted At")] = None
    deleted_by: User | None = None
    etag: Annotated[StrictStr, Field(title="Etag")]
    geojson_bounding_box: BoundingBox | None = None
    links: ListedObjectLinks
    modified_at: Annotated[datetime, Field(title="Modified At")]
    modified_by: User | None = None
    name: Annotated[StrictStr, Field(title="Name")]
    object_id: Annotated[UUID, Field(title="Object Id")]
    path: Annotated[StrictStr, Field(title="Path")]
    schema_: Annotated[StrictStr, Field(alias="schema", title="Schema")]
    """
    The geoscience object schema.
    """
    stage: StageResponse | None = None
    version_id: Annotated[StrictStr, Field(title="Version Id")]


class OrgListedObject(CustomBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: User | None = None
    deleted_at: Annotated[datetime | None, Field(title="Deleted At")] = None
    deleted_by: User | None = None
    geojson_bounding_box: BoundingBox | None = None
    modified_at: Annotated[datetime, Field(title="Modified At")]
    modified_by: User | None = None
    name: Annotated[StrictStr, Field(title="Name")]
    object_id: Annotated[UUID, Field(title="Object Id")]
    schema_: Annotated[StrictStr, Field(alias="schema", title="Schema")]
    stage: StageResponse | None = None
    workspace_access: RoleEnum | None = None
    workspace_id: Annotated[UUID, Field(title="Workspace Id")]
    workspace_name: Annotated[StrictStr | None, Field(title="Workspace Name")] = None


class PostObjectResponse(CustomBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    created_by: User | None = None
    deleted_at: Annotated[datetime | None, Field(title="Deleted At")] = None
    deleted_by: User | None = None
    etag: Annotated[StrictStr, Field(title="Etag")]
    geojson_bounding_box: BoundingBox | None = None
    links: ObjectResponseLinks
    modified_at: Annotated[datetime, Field(title="Modified At")]
    modified_by: User | None = None
    object: GeoscienceObject
    object_id: Annotated[UUID | None, Field(title="Object Id")] = None
    object_path: Annotated[StrictStr | None, Field(title="Object Path")] = None
    stage: StageResponse | None = None
    version_id: Annotated[StrictStr, Field(title="Version Id")]


class ListObjectsResponse(CustomBaseModel):
    count: Annotated[StrictInt, Field(title="Count")] = 0
    limit: Annotated[StrictInt, Field(title="Limit")] = 0
    links: ListObjectsResponseLinks
    objects: Annotated[list[ListedObject], Field(title="Objects")]
    offset: Annotated[StrictInt, Field(title="Offset")] = 0
    total: Annotated[StrictInt, Field(title="Total")] = 0


class ListOrgObjectsResponse(CustomBaseModel):
    count: Annotated[StrictInt, Field(title="Count")] = 0
    limit: Annotated[StrictInt, Field(title="Limit")] = 0
    links: ListObjectsResponseLinks
    objects: Annotated[list[OrgListedObject], Field(title="Objects")]
    offset: Annotated[StrictInt, Field(title="Offset")] = 0
    total: Annotated[StrictInt, Field(title="Total")] = 0



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Geoscience Object API
=============


The Geoscience Object API enables technological integrations at the object level. It enables users to access and utilise their data across all products through a common and accessible data structure.

A Geoscience Object is a data structure that represents a concrete geological, geotechnical, or geophysical concept. Geoscience Objects can be referenced by their UUID or by a user-defined object path.

For more information on using the Geoscience Object API, see the [Geoscience Object API overview](/docs/guides/objects), or the API references here.


This code is generated from the OpenAPI specification for Geoscience Object API.
API version: 1.14.0
"""

# Import endpoint apis.
from .api import DataApi, MetadataApi, ObjectsApi, StagesApi

__all__ = [
    "DataApi",
    "MetadataApi",
    "ObjectsApi",
    "StagesApi",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from evo.common.exceptions import CustomTypedError, EvoClientException


class SchemaIDFormatError(EvoClientException):
    """Unrecognized schema id format."""


class ObjectUUIDError(EvoClientException):
    """Object UUID is an unexpected value."""


class UnknownSchemaError(EvoClientException):
    """Unknown object schema."""

    def __init__(self, schema_id: str | None, detail: str):
        self.schema_id = schema_id
        self.detail = detail

    def __str__(self) -> str:
        return f"Could not resolve schema with id: '{self.schema_id}'\ndetail: {self.detail}"


class TableFormatError(EvoClientException):
    """The table structure does not match the expected Geoscience Object Schema data structure"""


class SchemaValidationError(EvoClientException):
    """Failed to validate data against a geoscience object schema"""


class ObjectAlreadyExistsError(CustomTypedError):
    TYPE_ID = "/geoscienceobject/objects/already-exists"

    @property
    def existing_id(self) -> UUID | None:
        if "existing_id" in self.content:
            return UUID(self.content["existing_id"])
        else:
            return None

    @property
    def object_path(self) -> str | None:
        if "object_path" in self.content:
            return str(self.content["object_path"])
        else:
            return None

    def __str__(self) -> str:
        error_message = super().__str__()
        if object_path := self.object_path:
            error_message += f"\nPath: {object_path}"
        if existing_id := self.existing_id:
            error_message += f"\nExisting ID: {existing_id}"
        return error_message
